from typing import Iterable, List, Union, Tuple

import numpy as np

from fdsreader.slcf import Slice
from fdsreader.utils.data import FDSDataCollection, Quantity


class SliceCollection(FDSDataCollection):
    """Collection of :class:`Slice` objects. Offers extensive functionality for filtering and
        using slices as well as its subclasses such as :class:`SubSlice`.
    """

    def __init__(self, *slices: Iterable[Slice]):
        super().__init__(*slices)

    def filter_by_quantity(self, quantity: Union[str, Quantity]):
        """Filters all slices by a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        print(SliceCollection(x for x in self))
        return SliceCollection(x for x in self if x.quantity.quantity.lower() == quantity.lower()
                               or x.quantity.label.lower() == quantity.lower())

    def get_by_id(self, slice_id: str):
        """Get the slice with corresponding id if it exists.
        """
        return next((slc for slc in self if slc.id == slice_id), None)

    def get_nearest(self, x: float = None, y: float = None, z: float = None) -> Slice:
        """Filters the slice with the shortest distance to the given point.
        """
        d_min = np.finfo(float).max
        slc_min = None

        for slc in self:
            dx = max(slc.extent.x_start - x, 0, x - slc.extent.x_end) if x is not None else 0
            dy = max(slc.extent.y_start - y, 0, y - slc.extent.y_end) if y is not None else 0
            dz = max(slc.extent.z_start - z, 0, z - slc.extent.z_end) if z is not None else 0
            d = np.sqrt(dx * dx + dy * dy + dz * dz)
            if d < d_min:
                d_min = d
                slc_min = slc
        return slc_min

    def __repr__(self):
        return "SliceCollection(" + super(SliceCollection, self).__repr__() + ")"



from typing import Optional, Tuple


class Surface:
    """Surface objects describe what bounding surfaces consist of. Boundary conditions for
        obstructions and vents are prescribed by referencing the appropriate surface.

    :ivar name: Name of the surface.
    :ivar material_emissivity: Emissivity of the material.
    :ivar surface_type: Type of the surface.
    :ivar texture_width:Width of the texture of the surface.
    :ivar texture_height: Height of the texture of the surface.
    :ivar texture_map: Path to the texture map used for the surface.
    :ivar rgb: Color of the surface in form of a 3-element tuple.
    :ivar transparency: Transparency of the color (alpha channel).
    """

    def __init__(self, name: str, tmpm: float, material_emissivity: float, surface_type: int,
                 texture_width: float, texture_height: float, texture_map: Optional[str],
                 rgb: Tuple[float, float, float], transparency: float):
        self.name = name
        self.tmpm = tmpm
        self.material_emissivity = material_emissivity
        self.surface_type = surface_type
        self.texture_width = texture_width
        self.texture_height = texture_height
        self.texture_map = texture_map
        self.rgb = rgb
        self.transparency = transparency

    def id(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self, *args, **kwargs):
        return f'Surface(name="{self.name}")'



import operator
from functools import reduce
from typing import BinaryIO, Dict, Union, List, Tuple, Optional

import numpy as np

from fdsreader.utils import Quantity, Mesh
from fdsreader import settings
import fdsreader.utils.fortran_data as fdtype


class SubSurface:
    """Part of an isosurface with data for a specific mesh.

    :ivar mesh: The mesh containing all data for this :class:`SubSurface`.
    :ivar file_path: Path to the binary data file.
    :ivar v_file_path: Path to the binary data file containing color data.
    :ivar n_vertices: The number of vertices for this subsurface.
    :ivar n_triangles: The number of triangles for this subsurface.
    :ivar n_t: Total number of time steps for which output data has been written.
    """

    def __init__(self, mesh: Mesh, iso_filepath: str, times: List, viso_filepath: str = ""):
        self.mesh = mesh

        self.file_path = iso_filepath
        if viso_filepath != "":
            self.v_file_path = viso_filepath

        with open(self.file_path, 'rb') as infile:
            nlevels = fdtype.read(infile, fdtype.INT, 3)[2][0][0]
            dtype_header_levels = fdtype.new((('f', nlevels),))
            dtype_header_zeros = fdtype.combine(fdtype.INT, fdtype.new((('i', 2),)))
            self._offset = fdtype.INT.itemsize * 3 + dtype_header_levels.itemsize + \
                           dtype_header_zeros.itemsize

            self.times = times
            self.n_vertices = list()
            self.n_triangles = list()

            if not settings.LAZY_LOAD:
                self._load_data(infile)

        if self.has_color_data:
            if not settings.LAZY_LOAD:
                with open(self.v_file_path, 'rb') as infile:
                    self._load_vdata(infile)

    @property
    def vertices(self):
        """Property to lazy load all vertices for all triangles of any level.
        """
        if not hasattr(self, "_vertices"):
            with open(self.file_path, 'rb') as infile:
                self._load_data(infile)
        return self._vertices

    @property
    def triangles(self):
        """Property to lazy load all triangles of any level.
        """
        if not hasattr(self, "_triangles"):
            with open(self.file_path, 'rb') as infile:
                self._load_data(infile)
        return self._triangles

    @property
    def surfaces(self):
        """Property to lazy load a list that maps triangles to an isosurface for a specific level.
        The list has the size n_triangles, while the indices correspond to indices of the triangles.
        """
        if not hasattr(self, "_surfaces"):
            with open(self.file_path, 'rb') as infile:
                self._load_data(infile)
        return self._surfaces

    @property
    def has_color_data(self):
        """Defines whether there is color data for this subsurface or not.
        """
        return hasattr(self, "v_file_path")

    @property
    def colors(self):
        """Property to lazy load the color data that might be associated with the isosurfaces.
        """
        if self.has_color_data:
            if not hasattr(self, "_colors"):
                with open(self.v_file_path, 'rb') as infile:
                    self._load_vdata(infile)
            return self._colors
        else:
            raise UserWarning("The isosurface does not have any associated color-data. Use the"
                              " attribute 'has_color_data' to check if an isosurface has associated"
                              " color-data.")

    def _load_data(self, infile: BinaryIO):
        """Loads data for the subsurface which is given in an iso file.
        """
        dtype_time = fdtype.new((('f', 1), ('i', 1)))
        dtype_dims = fdtype.new((('i', 2),))

        self._vertices = list()
        self._triangles = list()
        self._surfaces = list()

        infile.seek(self._offset)
        time_data = fdtype.read(infile, dtype_time, 1)
        while time_data.size != 0:
            self.times.append(time_data[0][0][0])

            dims_data = fdtype.read(infile, dtype_dims, 1)
            n_vertices = dims_data[0][0][0]
            n_triangles = dims_data[0][0][1]
            if n_vertices > 0:
                dtype_vertices = fdtype.new((('f', 3 * n_vertices),))
                dtype_triangles = fdtype.new((('i', 3 * n_triangles),))
                dtype_surfaces = fdtype.new((('i', n_triangles),))

                self._vertices.append(
                    fdtype.read(infile, dtype_vertices, 1)[0][0].reshape((n_vertices, 3)).astype(
                        float))
                self._triangles.append(
                    fdtype.read(infile, dtype_triangles, 1)[0][0].reshape((n_triangles, 3)).astype(
                        int) - 1)
                self._surfaces.append(fdtype.read(infile, dtype_surfaces, 1)[0][0].astype(int) - 1)
                self.n_vertices.append(n_vertices)
                self.n_triangles.append(n_triangles)
            else:
                self._vertices.append(np.empty((0, 3)))
                self._triangles.append(np.empty((0, 3)))
                self._surfaces.append(np.empty((0,)))
                self.n_vertices.append(0)
                self.n_triangles.append(0)

            time_data = fdtype.read(infile, dtype_time, 1)

        self.n_t = len(self.times)

    def _load_vdata(self, infile: BinaryIO):
        """Loads all color data for all isosurfaces in a given viso file.
        """
        self._colors = np.empty((self.n_t,), dtype=object)
        dtype_nverts = fdtype.new((('i', 4),))

        infile.seek(fdtype.INT.itemsize * 2)
        t = fdtype.read(infile, fdtype.FLOAT, 1)
        while t.size != 0:
            time_index = self.times.index(t[0][0][0])
            n_vertices = fdtype.read(infile, dtype_nverts, 1)[0][0][2]
            if n_vertices > 0:
                self._colors[time_index] = fdtype.read(infile, fdtype.new((('f', n_vertices),)), 1)
            t = fdtype.read(infile, fdtype.FLOAT, 1)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "times"):
            del self.times
        if hasattr(self, "_vertices"):
            del self._vertices
        if hasattr(self, "_triangles"):
            del self._triangles
        if hasattr(self, "_surfaces"):
            del self._surfaces
        if hasattr(self, "_colors"):
            del self._colors

    def __repr__(self):
        return f'SubSurface(mesh="{str(self.mesh)}")'


class Isosurface:
    """Isosurface file data container including metadata. Consists of a list of vertices forming a
        list of triangles. Can optionally have additional color data for the surfaces.

    :ivar id: The ID of this isosurface.
    :ivar quantity: Quantity object containing information about the quantity calculated for this
        isosurface with the corresponding label and unit.
    :ivar v_quantity: Information about the color quantity.
    :ivar levels: All isosurface levels.
    """

    def __init__(self, isosurface_id: int, double_quantity: bool, quantity: str,
                 label: str, unit: str, levels: List[float], v_quantity: str = "",
                 v_label: str = "", v_unit: str = ""):
        self.id = isosurface_id
        self.quantity = Quantity(quantity, label, unit)
        self._double_quantity = double_quantity
        self.levels = levels

        self._times = list()

        self._subsurfaces: Dict[Mesh, SubSurface] = dict()

        if self._double_quantity:
            self.v_quantity = Quantity(v_quantity, v_label, v_unit)

    def _add_subsurface(self, mesh: Mesh, iso_file_path: str,
                        viso_file_path: str = "") -> SubSurface:
        if viso_file_path != "":
            subsurface = SubSurface(mesh, iso_file_path, self._times, viso_file_path)
        else:
            subsurface = SubSurface(mesh, iso_file_path, self._times)
        self._subsurfaces[mesh] = subsurface

        return subsurface

    def __getitem__(self, mesh: Mesh) -> SubSurface:
        """Returns the :class:`SubSurface` that contains data for the given mesh.
        """
        return self._subsurfaces[mesh]

    @property
    def vertices(self) -> Dict[Mesh, List[np.ndarray]]:
        """Gets all vertices per mesh.
        """
        return {mesh: subsurface.vertices for mesh, subsurface in self._subsurfaces.items()}

    @property
    def triangles(self) -> Dict[Mesh, List[np.ndarray]]:
        """Gets all triangles per mesh.
        """
        return {mesh: subsurface.triangles for mesh, subsurface in self._subsurfaces.items()}

    @property
    def surfaces(self) -> Dict[Mesh, List[np.ndarray]]:
        """Gets all surfaces per mesh.
        """
        return {mesh: subsurface.surfaces for mesh, subsurface in self._subsurfaces.items()}

    def to_global(self, time: Union[int, float]) -> Tuple[
        np.ndarray, List[np.ndarray], Optional[np.ndarray]]:
        """Creates an array containing all global vertices and a list containing numpy arrays with
            triangles for each surface level.

        :param time: Either the index of the timestep or an actual time value. In the latter case
            data for the nearest matching timestep will be used.
        """
        if type(time) == float:
            time = self.get_nearest_timestep(time)

        if time > len(self.times):
            time = len(self.times) - 1
        if time < 0:
            time = 0

        n_vertices = sum(
            x[time].shape[0] if len(x[time].shape) > 0 else 0 for x in self.vertices.values())
        vertices = np.empty((n_vertices, 3))

        colors = np.empty((n_vertices,)) if self.has_color_data else None
        verts_counter = 0
        for mesh in self._subsurfaces.keys():
            tmp = verts_counter
            verts = self[mesh].vertices[time]
            verts_counter += verts.shape[0]
            vertices[tmp:verts_counter] = verts
            if self.has_color_data:
                colors[tmp: verts_counter] = self[mesh].colors[time]

        triangles = list()
        num_levels = max(
            max(np.max(t) if t.shape[0] != 0 else 0 for t in s) for s in self.surfaces.values()) + 1
        verts_counter = 0
        for surf in range(num_levels):
            n_triangles = sum(np.count_nonzero(s[time] == surf) for s in self.surfaces.values())
            triangles.append(np.empty((n_triangles, 3), dtype=int))
            tris_counter = 0
            if n_triangles > 0:
                for mesh in self._subsurfaces.keys():
                    tmp = tris_counter
                    tris = self.triangles[mesh][time][self.surfaces[mesh][time] == surf]
                    tris_counter += tris.shape[0]
                    triangles[surf][tmp:tris_counter] = tris + verts_counter
                    verts_counter += self[mesh].vertices[time].shape[0]

        return vertices, triangles, colors

    def get_pyvista_mesh(self, vertices: np.ndarray, triangles: np.ndarray):
        """Creates a PyVista mesh from the data.
        """
        try:
            from pyvista import PolyData

            triangles = np.hstack(np.append(np.full((triangles.shape[0], 1), 3), triangles, axis=1))
            return PolyData(vertices, triangles)
        except ImportError:
            raise ImportError("The 'get_pyvista_mesh' method requires the PyVista python-package to"
                              " be installed. Consider installing it via 'pip install pyvista'.")

    def join_pyvista_meshes(self, meshes: List):
        """Combines multiple PyVista meshes.

        :returns: The combined mesh of class PolyData.
        """
        try:
            from pyvista import PolyData

            assert all(type(mesh) == PolyData for mesh in
                       meshes), "Argument 'meshes' has to be a sequence of type pyvista.Polydata"

            return reduce(operator.add, meshes)
        except ImportError:
            raise ImportError("The 'join_pyvista_meshes' method requires the PyVista python-package"
                              " to be installed. Consider installing it via 'pip install pyvista'.")

    def export(self, file_path: str, mesh):
        """Export the isosurface for a single timestep into one of many formats.

        :param file_path: Absolute path to the the file which should we written. The file ending
            denotes the file format to export to (supports pretty much every common format).
        :param time: Either the index of the timestep or an actual time value. In the latter case
            data for the nearest matching timestep will be used.
        """
        try:
            from pyvista import PolyData

            assert type(mesh) == PolyData, "Argument 'mesh' has to be of type pyvista.Polydata"

            if "vtk" in file_path or "vtp" in file_path:
                return mesh.save(file_path)
            else:
                try:
                    import meshio

                    return mesh.save_meshio(file_path)
                except ImportError:
                    raise ImportError("The 'export' method requires the Meshio python-package to be"
                                      " installed. "
                                      "Consider installing it via 'pip install meshio'.")

        except ImportError:
            raise ImportError("The 'export' method requires the PyVista python-package to be"
                              " installed. Consider installing it via 'pip install pyvista'.")

    def get_nearest_timestep(self, time: float) -> int:
        """Calculates the nearest timestep for which data has been output for this isosurface.
        """
        idx = np.searchsorted(self.times, time, side="left")
        if time > 0 and (idx == len(self.times) or np.math.fabs(
                time - self.times[idx - 1]) < np.math.fabs(time - self.times[idx])):
            return idx - 1
        else:
            return idx

    @property
    def has_color_data(self) -> bool:
        """Defines whether there is color data for this isosurface or not.
        """
        return self._double_quantity

    @property
    def times(self) -> List[float]:
        """List containing all times for which data has been recorded.
        """
        if len(self._times) == 0:
            # Implicitly load the data for one subsurface and read times
            _ = next(iter(self._subsurfaces.values())).vertices
        return self._times

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for subsurface in self._subsurfaces.values():
            subsurface.clear_cache()

    def __repr__(self):
        return f'Isosurface(id="{self.id}", quantity={str(self.quantity)}, levels={str(self.levels)})'



"""
Provides fdsreader version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update fdsreader` to change this file.

from incremental import Version

__version__ = Version("fdsreader", 0, 9, 15)
__all__ = ["__version__"]



import os
from copy import deepcopy, copy

import numpy as np
import logging
from typing import Dict, Collection, Tuple, Union, List
from typing_extensions import Literal

from fdsreader.utils import Dimension, Quantity, Mesh, Extent
from fdsreader import settings
import fdsreader.utils.fortran_data as fdtype

_HANDLED_FUNCTIONS = {}


def implements(np_function):
    """Decorator to register an __array_function__ implementation for Slices.
    """

    def decorator(func):
        _HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


class SubSlice:
    """Part of a slice that cuts through a single mesh.

    :ivar mesh: The mesh the subslice cuts through.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar dimension: :class:`Dimension` object containing information about steps in each dimension.
    """

    _offset = 3 * fdtype.new((('c', 30),)).itemsize + fdtype.new((('i', 6),)).itemsize

    def __init__(self, parent_slc, filename: str, dimension: Dimension, extent: Extent, mesh: Mesh):
        self.mesh = mesh
        self.dimension = dimension
        self.extent = extent
        self.parent_slice = parent_slc

        self.filename = filename

        self.vector_filenames = dict()
        self._vector_data = dict()

    @property
    def shape(self) -> Tuple[int, int]:
        """2D-shape of the slice.
        """
        shape = self.dimension.shape(cell_centered=self.parent_slice.cell_centered)
        if self.parent_slice.orientation != 0:
            return shape[0], shape[1]
        return shape

    @property
    def orientation(self):
        """Orientation [1,2,3] of the slice in case it is 2D, 0 otherwise.
        """
        return self.parent_slice.orientation

    def _load_data(self, file_path: str, data_out: np.ndarray):
        # Both cases (cell_centered True/False) output the same number of data points
        n = self.dimension.size(cell_centered=False)
        dtype_data = fdtype.combine(fdtype.FLOAT, fdtype.new((('f', n),)))

        load_times = False
        if self.parent_slice.n_t == -1:
            load_times = True
            self.parent_slice.n_t = (os.stat(
                file_path).st_size - self._offset) // dtype_data.itemsize
            self.parent_slice.times = np.empty(self.parent_slice.n_t)

        with open(file_path, 'rb') as infile:
            infile.seek(self._offset)
            for t, data in enumerate(fdtype.read(infile, dtype_data, self.parent_slice.n_t)):
                if load_times:
                    self.parent_slice.times[t] = data[0][0]
                data = data[1].reshape(self.dimension.shape(cell_centered=False), order='F')
                if self.parent_slice.cell_centered:
                    data_out[t, :] = data[:-1, :-1]  # Ignore ghost points
                else:
                    data_out[t, :] = data

    @property
    def data(self) -> np.ndarray:
        """Method to lazy load the slice's data.
        """
        if not hasattr(self, "_data"):
            file_path = os.path.join(self.parent_slice.root_path, self.filename)
            self._data = np.empty((self.parent_slice.n_t,) + self.shape, dtype=np.float32)
            self._load_data(file_path, self._data)
        return self._data

    @property
    def vector_data(self) -> Dict[str, np.ndarray]:
        """Method to lazy load the slice's vector data if it exists.
        """
        if not hasattr(self, "_vector_data"):
            raise AttributeError("There is no vector data available for this slice.")
        if len(self._vector_data) == 0:
            for direction in self.vector_filenames.keys():
                file_path = os.path.join(self.parent_slice.root_path,
                                         self.vector_filenames[direction])
                self._vector_data[direction] = np.empty((self.parent_slice.n_t,) + self.shape,
                                                        dtype=np.float32)
                self._load_data(file_path, self._vector_data[direction])
        return self._vector_data

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data
        if hasattr(self, "_vector_data"):
            del self._vector_data

    def __repr__(self):
        return f"SubSlice(shape={self.shape}, mesh={self.mesh.id}, extent={self.extent})"


class Slice(np.lib.mixins.NDArrayOperatorsMixin):
    """Slice file data container including metadata. Consists of multiple subslices, one for each
        mesh the slice cuts through.

    :ivar root_path: Path to the directory containing all slice files.
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar quantity: Quantity object containing information about the quantity calculated for this
        slice with the corresponding label and unit.
    :ivar times: Numpy array containing all times for which data has been recorded.
    :ivar n_t: Total number of time steps for which output data has been written.
    :ivar orientation: Orientation [1,2,3] of the slice in case it is 2D, 0 otherwise.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    """

    def __init__(self, root_path: str, slice_id: str, cell_centered: bool, times: np.ndarray,
                 multimesh_data: Collection[Dict]):
        self.root_path = root_path
        self.cell_centered = cell_centered

        self.times = times

        if times is not None:
            self.n_t = times.size
        else:
            self.n_t = -1

        self.id = slice_id

        # List of all subslices this slice consists of (one per mesh).
        self._subslices: Dict[Mesh, SubSlice] = dict()

        vector_temp = dict()
        for mesh_data in multimesh_data:
            if "-VELOCITY" in mesh_data["quantity"]:
                vector_temp[mesh_data["mesh"]] = dict()

        for mesh_data in multimesh_data:
            if mesh_data["mesh"] not in self._subslices:
                self.quantity = Quantity(mesh_data["quantity"], mesh_data["label"],
                                         mesh_data["unit"])
                self._subslices[mesh_data["mesh"]] = SubSlice(self, mesh_data["filename"],
                                                              mesh_data["dimension"],
                                                              mesh_data["extent"],
                                                              mesh_data["mesh"])

            if "-VELOCITY" in mesh_data["quantity"]:
                vector_temp[mesh_data["mesh"]][mesh_data["quantity"]] = mesh_data["filename"]

        for mesh, vector_filenames in vector_temp.items():
            if "U-VELOCITY" in vector_filenames:
                self._subslices[mesh].vector_filenames["u"] = vector_filenames["U-VELOCITY"]
            if "V-VELOCITY" in vector_filenames:
                self._subslices[mesh].vector_filenames["v"] = vector_filenames["V-VELOCITY"]
            if "W-VELOCITY" in vector_filenames:
                self._subslices[mesh].vector_filenames["w"] = vector_filenames["W-VELOCITY"]

        # Iterate over all subslices and remove duplicated ones which will be created when the slice
        # cuts exactly through two mesh borders. Only required for non-cell-centered slices.
        if not self.cell_centered:
            extents_tmp = list()
            remove_tmp = list()
            for mesh, sslc in self._subslices.items():
                if sslc.extent not in extents_tmp:
                    extents_tmp.append(sslc.extent)
                else:
                    remove_tmp.append(mesh)
            for mesh in remove_tmp:
                del self._subslices[mesh]

        vals = self._subslices.values()
        self.extent = Extent(min(vals, key=lambda e: e.extent.x_start).extent.x_start,
                             max(vals, key=lambda e: e.extent.x_end).extent.x_end,
                             min(vals, key=lambda e: e.extent.y_start).extent.y_start,
                             max(vals, key=lambda e: e.extent.y_end).extent.y_end,
                             min(vals, key=lambda e: e.extent.z_start).extent.z_start,
                             max(vals, key=lambda e: e.extent.z_end).extent.z_end)

        if self.extent.x_start == self.extent.x_end:
            self.orientation = 1
        elif self.extent.y_start == self.extent.y_end:
            self.orientation = 2
        elif self.extent.z_start == self.extent.z_end:
            self.orientation = 3
        else:
            self.orientation = 0

        # If lazy loading has been disabled by the user, load the data instantaneously instead
        if not settings.LAZY_LOAD:
            for _, subslice in self._subslices.items():
                _ = subslice.data

    def __getitem__(self, key: Union[int, Mesh]):
        """Returns the :class:`SubSlice` that cuts through the given mesh.
        """
        if type(key) == int:
            return tuple(self._subslices.values())[key]
        return self._subslices[key]

    def __len__(self):
        return len(self._subslices)

    @property
    def extent_dirs(self) -> Tuple[
        Literal['x', 'y', 'z'], Literal['x', 'y', 'z'], Literal['x', 'y', 'z']]:
        """The directions in which there is an extent. All three dimensions in case the slice is 3D.
        """
        ior = self.orientation
        if ior == 0:
            return 'x', 'y', 'z'
        elif ior == 1:
            return 'y', 'z'
        elif ior == 2:
            return 'x', 'z'
        else:
            return 'x', 'y'

    def get_nearest_timestep(self, time: float) -> int:
        """Calculates the nearest timestep for which data has been output for this slice.
        """
        idx = np.searchsorted(self.times, time, side="left")
        if time > 0 and (idx == len(self.times) or np.math.fabs(
                time - self.times[idx - 1]) < np.math.fabs(time - self.times[idx])):
            return idx - 1
        else:
            return idx

    def get_nearest_index(self, dimension: Literal['x', 'y', 'z'], value: float) -> int:
        coords = self.coordinates[dimension]
        idx = np.searchsorted(coords, value, side="left")
        if idx > 0 and (idx == coords.size or np.math.fabs(value - coords[idx - 1]) < np.math.fabs(
                value - coords[idx])):
            return idx - 1
        else:
            return idx

    @property
    def meshes(self) -> List[Mesh]:
        """Returns a list of all meshes this slice cuts through.
        """
        return list(self._subslices.keys())

    @property
    def coordinates(self) -> Dict[Literal['x', 'y', 'z'], np.ndarray]:
        coords = {'x': set(), 'y': set(), 'z': set()}
        for dim in ('x', 'y', 'z'):
            for mesh in self._subslices.keys():
                coords[dim].update(mesh.coordinates[dim])
            coords[dim] = np.array(sorted(list(coords[dim])))
        return coords

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for subslice in self._subslices.values():
            subslice.clear_cache()

    def sort_subslices_cartesian(self):
        """Returns all subslices sorted in cartesian coordinates (2D-slices only).
        """
        assert self.type == '2D', "The sort_subslices_cartesian method only works on 2D-slices!"

        slices = list(self._subslices.values())
        slices_cart = [[slices[0]]]
        orientation = abs(slices[0].orientation)
        if orientation == 1:  # x
            slices.sort(key=lambda p: (p.extent.y_start, p.extent.z_start))
        elif orientation == 2:  # y
            slices.sort(key=lambda p: (p.extent.x_start, p.extent.z_start))
        elif orientation == 3:  # z
            slices.sort(key=lambda p: (p.extent.x_start, p.extent.y_start))

        if orientation == 1:
            for slc in slices[1:]:
                if slc.extent.y_start == slices_cart[-1][-1].extent.y_start:
                    slices_cart[-1].append(slc)
                else:
                    slices_cart.append([slc])
        else:
            for slc in slices[1:]:
                if slc.extent.x_start == slices_cart[-1][-1].extent.x_start:
                    slices_cart[-1].append(slc)
                else:
                    slices_cart.append([slc])
        return slices_cart

    def to_global(self) -> np.ndarray:
        """Creates a global 2D-numpy ndarray from all subslices (2D-slices only).
            Note: This method will only return valid output if evenly sized and spaced meshes were
            used in the simulation. Other cases will require individual custom combination logic.
        """
        assert self.type == '2D', "The to_global method only works on 2D-slices!"

        slices = self.sort_subslices_cartesian()

        shape_dim1 = sum([slice_row[0].shape[0] for slice_row in slices])
        if not self.cell_centered:
            shape_dim1 -= len(slices) - 1
        shape_dim2 = sum([slc.shape[1] for slc in slices[0]])
        if not self.cell_centered:
            shape_dim2 -= len(slices[0]) - 1
        slc_array = np.empty(shape=(self.n_t, shape_dim1, shape_dim2))
        dim1_pos = 0
        dim2_pos = 0
        for i, slice_row in enumerate(slices):
            d1 = slice_row[0].shape[0]
            if not self.cell_centered:
                d1 -= 1 if i + 1 != len(slices) else 0
            for j, slc in enumerate(slice_row):
                d2 = slc.shape[1]
                if not self.cell_centered:
                    d2 -= 1 if j + 1 != len(slice_row) else 0
                slc_array[:, dim1_pos:dim1_pos + d1, dim2_pos:dim2_pos + d2] = slc.data[:, :d1, :d2]
                dim2_pos += d2
            dim1_pos += d1
            dim2_pos = 0
        return slc_array

    @property
    def type(self) -> Literal['2D', '3D']:
        if self.orientation == 0:
            return '3D'
        return '2D'

    def mean(self):
        """Calculates the mean over the whole slice.

        :returns: The calculated mean value.
        """
        return np.mean([np.mean(subsclice.data) for subsclice in self._subslices.values()])

    def __array__(self):
        """Method that will be called by numpy when trying to convert the object to a numpy ndarray.
        """
        raise UserWarning(
            "Slices can not be converted to numpy arrays, but they support all typical numpy"
            " operations such as np.multiply. If a 'global' array containg all subslices is"
            " required, use the 'to_global' method and use the returned numpy-array explicitly.")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Method that will be called by numpy when using a ufunction with a Slice as input.

        :returns: A new slice on which the ufunc has been applied.
        """
        if method != "__call__":
            logging.warning(
                "The %s method has been used which is not explicitly implemented. Correctness of"
                " results is not guaranteed. If you require this feature to be implemented please"
                " submit an issue on Github where you explain your use case.", method)
        input_list = list(inputs)
        for i, inp in enumerate(inputs):
            if isinstance(inp, self.__class__):
                del input_list[i]
        if len(input_list) == 0:
            raise UserWarning(
                f"The {method} operation is not implemented for multiple slices as input yet. If"
                " you require this feature, please request this functionality by submitting an"
                " issue on Github.")

        new_slice = deepcopy(self)
        for subslice in new_slice._subslices.values():
            subslice._data = ufunc(subslice.data, input_list[0], **kwargs)
        return new_slice

    def __array_function__(self, func, types, args, kwargs):
        """Method that will be called by numpy when using an array function with a Slice as input.

        :returns: The output of the array function.
        """
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
            # Note: this allows subclasses that don't override __array_function__ to handle Slices.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __repr__(self):
        if self.type == '3D':  # 3D-Slice
            return f"Slice([3D] cell_centered={self.cell_centered}, extent={self.extent}, times=[{self.times[0]:.2f},{self.times[1]:.2f},...,{self.times[-1]:.2f}])"
        else:  # 2D-Slice
            return f"Slice([2D] cell_centered={self.cell_centered}, extent={self.extent}, extent_dirs={self.extent_dirs}, orientation={self.orientation}, times=[{self.times[0]:.2f},{self.times[1]:.2f},...,{self.times[-1]:.2f}])"

# __array_function__ implementations



from typing import Tuple, List

from typing_extensions import Literal


class Extent:
    """Three-dimensional value-based extent with support for a missing dimension (2D).
    """

    def __init__(self, *args, skip_dimension: Literal['x', 1, 'y', 2, 'z', 3, ''] = ''):
        self._extents = list()

        if len(args) % 2 != 0:
            ValueError("An invalid number of arguments were passed to the constructor.")
        for i in range(0, len(args), 2):
            self._extents.append((float(args[i]), float(args[i + 1])))

        if skip_dimension in ('x', 1):
            self._extents.insert(0, (0, 0))
        elif skip_dimension in ('y', 2):
            self._extents.insert(1, (0, 0))
        elif skip_dimension in ('z', 3):
            self._extents.append((0, 0))

    def __eq__(self, other):
        return self._extents == other._extents

    def __repr__(self, *args, **kwargs):
        return "Extent([{:.2f}, {:.2f}] x [{:.2f}, {:.2f}] x [{:.2f}, {:.2f}])".format(self.x_start,
                                                                                       self.x_end,
                                                                                       self.y_start,
                                                                                       self.y_end,
                                                                                       self.z_start,
                                                                                       self.z_end)

    @property
    def x_start(self):
        """Gives the absolute extent in x-direction.
        """
        return self._extents[0][0]

    @property
    def y_start(self):
        """Gives the absolute extent in y-direction.
        """
        return self._extents[1][0]

    @property
    def z_start(self):
        """Gives the absolute extent in z-direction.
        """
        return self._extents[2][0]

    @property
    def x_end(self):
        """Gives the absolute extent in x-direction.
        """
        return self._extents[0][1]

    @property
    def y_end(self):
        """Gives the absolute extent in y-direction.
        """
        return self._extents[1][1]

    @property
    def z_end(self):
        """Gives the absolute extent in z-direction.
        """
        return self._extents[2][1]

    def as_tuple(self, reduced=True) -> Tuple:
        """Gives the extent in tuple notation (without empty extents).

        :param reduced: Whether to leave out empty extents or not.
        """
        if reduced:
            if self.x_start == self.x_end:
                return self.y_start, self.y_end, self.z_start, self.z_end
            elif self.y_start == self.y_end:
                return self.x_start, self.x_end, self.z_start, self.z_end
            elif self.z_start == self.z_end:
                return self.x_start, self.x_end, self.y_start, self.y_end
        return self.x_start, self.x_end, self.y_start, self.y_end, self.z_start, self.z_end

    def as_list(self, reduced=True) -> List:
        """Gives the extent in list notation (without empty extents).

        :param reduced: Whether to leave out empty extents or not.
        """
        return list(self.as_tuple(reduced))



from functools import reduce
from operator import mul
from typing import Tuple, List

from typing_extensions import Literal


class Dimension:
    """Three-dimensional index-based extent with support for a missing dimension (2D).

    :ivar x: Number of data points in x-direction (end is exclusive).
    :ivar y: Number of data points in y-direction (end is exclusive).
    :ivar z: Number of data points in z-direction (end is exclusive).
    """

    def __init__(self, *args, skip_dimension: Literal['x', 1, 'y', 2, 'z', 3, ''] = ''):
        dimensions = list(args)

        if skip_dimension in ('x', 1):
            dimensions.insert(0, 1)
        elif skip_dimension in ('y', 2):
            dimensions.insert(1, 1)
        elif skip_dimension in ('z', 3):
            dimensions.append(1)

        self.x = dimensions[0]
        self.y = dimensions[1]
        self.z = dimensions[2]

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __repr__(self, *args, **kwargs):
        return f"Dimension({self.x}, {self.y}, {self.z})"

    def __getitem__(self, dimension: Literal[0, 1, 2, 'x', 'y', 'z']):
        if type(dimension) == int:
            dimension = ('x', 'y', 'z')[dimension]
        return self.__dict__[dimension]

    def size(self, cell_centered=False):
        return reduce(mul, self.shape(cell_centered))

    def shape(self, cell_centered=False) -> Tuple:
        """Method to get the actual number of data points per dimension.
        """
        s = list()
        c = -1 if cell_centered else 0
        if self.x != 1:
            s.append(self.x + c)
        if self.y != 1:
            s.append(self.y + c)
        if self.z != 1:
            s.append(self.z + c)
        return tuple(s)

    def as_tuple(self, reduced=True) -> Tuple:
        """Gives the dimensions in tuple notation (optionally without empty dimensions).

        :param reduced: Whether to leave out empty dimensions (size of 1) or not.
        """
        if reduced:
            if self.x == 1:
                return self.y, self.z
            elif self.y == 1:
                return self.x, self.z
            elif self.z == 1:
                return self.x, self.y
        return self.x, self.y, self.z

    def as_list(self, reduced=True) -> List:
        """Gives the dimension in list notation (without empty extents).

        :param reduced: Whether to leave out empty dimensions (size of 1) or not.
        """
        return list(self.as_tuple(reduced))



"""
Collection of internal utilities (convenience functions) for data handling.
"""
import glob
import hashlib
import os
from collections import Iterable
from typing import Tuple
import numpy as np


class Quantity:
    """Object containing information about a quantity with the corresponding label and unit.
    """
    def __init__(self, quantity: str, label: str, unit: str):
        self.label = label
        self.unit = unit
        self.quantity = quantity

    def __eq__(self, other):
        return self.quantity == other.quantity

    def __hash__(self):
        return hash(self.label)

    def __str__(self):
        return f"Quantity(label={self.label}, unit={self.unit}, quantity={self.quantity})"

    def __repr__(self):
        return f"Quantity('{self.label}')"


class Device:
    """Represents a single Device.
    """
    def __init__(self, quantity: Quantity, position: Tuple[float, float, float],
                 orientation: Tuple[float, float, float]):
        self.quantity = quantity
        self.position = position
        self.orientation = orientation
        self.data: np.ndarray = None

    @property
    def name(self):
        return self.quantity.quantity

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return f"Device(name={self.name}, position={self.position}, quantity={self.quantity}, mean={np.mean(self.data)})"


def create_hash(path: str):
    """Returns the md5 hash as string for the given file.
    """
    return str(hashlib.md5(open(path, 'rb').read()))


def scan_directory_smv(directory: str):
    """Scanning a directory non-recursively for smv-files.

    :param directory: The directory that will be scanned for smv files.
    :returns: A list containing the path to each smv-file found in the directory.
    """
    return glob.glob(directory + "/**/*.smv", recursive=True)


def get_smv_file(path: str):
    """Get the .smv file in a given directory.

    :param path: Either the path to the directory containing the simulation data or direct path
        to the .smv file for the simulation in case that multiple simulation output was written to
        the same directory.
    """
    if os.path.isfile(path):
        return path
    elif os.path.isdir(path):
        files = scan_directory_smv(path)
        if len(files) > 1:
            raise IOError("There are multiple simulations in this path!")
        elif len(files) == 0:
            raise IOError("There are no simulations in this path!")
        return files[0]
    elif os.path.isfile(path + ".smv"):
        return path + ".smv"
    else:
        raise IOError("Path is invalid!")


class FDSDataCollection:
    """(Abstract) Base class for any collection of FDS data.
    """

    def __init__(self, *elements: Iterable):
        self._elements = tuple(*elements)

    def __getitem__(self, index):
        return self._elements[index]

    def __iter__(self):
        return self._elements.__iter__()

    def __len__(self):
        return len(self._elements)

    def __contains__(self, value):
        return value in self._elements

    def __repr__(self):
        return "[" + str(self._elements) + "]"

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for element in self._elements:
            element.clear_cache()



from typing import Iterable, Union, List

from fdsreader.isof import Isosurface
from fdsreader.utils.data import FDSDataCollection, Quantity


class IsosurfaceCollection(FDSDataCollection):
    """Collection of :class:`Isosurface` objects. Offers extensive functionality for filtering and
        using isosurfaces as well as its subclasses such as :class:`SubSurface`.
    """

    def __init__(self, *isosurfaces: Iterable[Isosurface]):
        super().__init__(*isosurfaces)

    def filter_by_quantity(self, quantity: Union[str, Quantity]):
        """Filters all isosurfaces by a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        return IsosurfaceCollection(x for x in self if
                                    x.quantity.quantity.lower() == quantity.lower() or x.quantity.label.lower() == quantity.lower())

    def __repr__(self):
        return "IsosurfaceCollection(" + super(IsosurfaceCollection, self).__repr__() + ")"



from typing import Iterable, Union, List

from fdsreader import settings
from fdsreader.geom import GeomBoundary
from fdsreader.utils.data import FDSDataCollection, Quantity


class GeometryCollection(FDSDataCollection):
    """Collection of :class:`GeomBoundary` objects. Offers extensive functionality for filtering and
        using geometry data.
    """

    def __init__(self, *geom_boundaries: Iterable[GeomBoundary]):
        super().__init__(*geom_boundaries)

        if not settings.LAZY_LOAD:
            for geom in self:
                geom._load_data()

    def filter_by_quantity(self, quantity: Union[str, Quantity]):
        """Filters all GeomBoundaries by a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        return GeometryCollection(x for x in self if
                                  x.quantity.quantity.lower() == quantity.lower() or x.quantity.label.lower() == quantity.lower())

    def __repr__(self):
        return "GeometryCollection(" + super(GeometryCollection, self).__repr__() + ")"



from typing import List, Tuple, Dict

import numpy as np

from fdsreader.utils import Quantity, Mesh


class Particle:
    """Container to store particle data from particle simulations with FDS.

    :ivar class_name: Name of the particle class defined in the FDS input-file.
    :ivar quantities: List of all quantities for which data has been written out.
    :ivar color: Color assigned to the particle.
    :ivar n_particles: Number of existing particles for each timestep per mesh.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep with quantities as keys.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep with quantities as keys.
    """

    def __init__(self, class_name: str, quantities: List[Quantity], color: Tuple[float, float, float]):
        self.class_name = class_name
        self.quantities = quantities
        self.color = color
        self.n_particles: Dict[Mesh, List[int]] = dict()

        self._positions = None
        self._tags = None
        self._data = {q.quantity: [] for q in self.quantities}

        self.lower_bounds = {q.quantity: [] for q in self.quantities}
        self.upper_bounds = {q.quantity: [] for q in self.quantities}

        self._init_callback = None

    @property
    def id(self):
        return self.class_name

    @property
    def data(self) -> Dict[str, List[np.ndarray]]:
        """Dictionary with quantities as keys and a list with a numpy array for each timestep which
            contains data for each particle in that timestep.
        """
        if self._positions is None:
            self._init_callback()
        return self._data

    @property
    def tags(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains a tag for each particle in that
            timestep.
        """
        if self._positions is None:
            self._init_callback()
        return self._tags

    @property
    def positions(self) -> List[np.ndarray]:
        """List with a numpy array for each timestep which contains the position of each particle in
            that timestep.
        """
        if self._positions is None:
            self._init_callback()
        return self._positions

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        del self._data

    def __repr__(self):
        return f"Particle(name={self.class_name}, quantities={self.quantities})"



from typing import Iterable, Dict
import numpy as np

from fdsreader.part import Particle
from fdsreader.utils import Mesh
from fdsreader.utils.data import FDSDataCollection
import fdsreader.utils.fortran_data as fdtype
from fdsreader import settings


class ParticleCollection(FDSDataCollection):
    """Collection of :class:`Plot3D` objects. Offers extensive functionality for filtering and
        using plot3Ds as well as its subclasses such as :class:`SubPlot3D`.
    """

    def __init__(self, times: Iterable[float], *particles: Iterable[Particle]):
        super().__init__(*particles)
        self.times = list(times)
        self._file_paths: Dict[Mesh, str] = dict()

        if settings.LAZY_LOAD:
            for particle in self:
                particle._init_callback = self._load_data

    def _post_init(self):
        if not settings.LAZY_LOAD:
            print("Test")
            self._load_data()

    def _load_data(self):
        """Function to read in all particle data for a simulation.
        """
        particles = self
        pointer_location = {particle: [0] * len(self.times) for particle in particles}

        for particle in particles:
            if particle._positions is None:
                particle._positions = list()
                particle._tags = list()
                for t in range(len(self.times)):
                    size = 0
                    for mesh in self._file_paths.keys():
                        size += particle.n_particles[mesh][t]
                    for quantity in particle.quantities:
                        particle._data[quantity.quantity].append(np.empty((size,)))
                    particle._positions.append(np.empty((size, 3)))
                    particle._tags.append(np.empty((size,)))

        for mesh, file_path in self._file_paths.items():
            with open(file_path, 'rb') as infile:
                # Initial offset (ONE, fds version and number of particle classes)
                offset = 3 * fdtype.INT.itemsize
                # Number of quantities for each particle class (plus an INTEGER_ZERO)
                offset += fdtype.new((('i', 2),)).itemsize * len(particles)
                # 30-char long name and unit information for each quantity
                offset += fdtype.new((('c', 30),)).itemsize * 2 * sum(
                    [len(particle.quantities) for particle in particles])
                infile.seek(offset)

                for t in range(len(self.times)):
                    # Skip time value
                    infile.seek(fdtype.FLOAT.itemsize, 1)

                    # Read data for each particle class
                    for particle in particles:
                        # Read number of particles in each class
                        n_particles = fdtype.read(infile, fdtype.INT, 1)[0][0][0]

                        offset = pointer_location[particle][t]
                        # Read positions
                        dtype_positions = fdtype.new((('f', 3 * n_particles),))
                        particle._positions[t][offset: offset + n_particles] = \
                            fdtype.read(infile, dtype_positions, 1)[0][0].reshape(
                                (n_particles, 3), order='F').astype(float)

                        # Read tags
                        dtype_tags = fdtype.new((('i', n_particles),))
                        particle._tags[t][offset: offset + n_particles] = \
                        fdtype.read(infile, dtype_tags, 1)[0][0]

                        # Read actual quantity values
                        dtype_data = fdtype.new(
                            (('f', str((n_particles, len(particle.quantities)))),))
                        data_raw = fdtype.read(infile, dtype_data, 1)[0][0].reshape(
                            (n_particles, len(particle.quantities)), order='F')

                        for q, quantity in enumerate(particle.quantities):
                            particle._data[quantity.quantity][t][
                            offset:offset + n_particles] = data_raw[:, q].astype(float)
                        pointer_location[particle][t] += particle.n_particles[mesh][t]

    def __getitem__(self, key):
        if type(key) == int:
            return self._elements[key]
        for particle in self:
            if particle.class_name == key:
                return particle

    def __contains__(self, value):
        if value in self._elements:
            return True
        for particle in self:
            if particle.class_name == value:
                return True
        return False

    def __repr__(self):
        return "ParticleCollection(" + super(ParticleCollection, self).__repr__() + ")"



from typing import Tuple, Dict, Iterable

import numpy as np

from fdsreader.utils import Surface, Quantity
import fdsreader.utils.fortran_data as fdtype


class GeomBoundary:
    """Boundary data of a specific quantity for all geoms in the simulation.

    :ivar quantity: Quantity object containing information about the quantity calculated for geoms.
    :ivar times: Numpy array containing all times for which data has been recorded.
    :ivar n_t: Total number of time steps for which output data has been written.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep per mesh.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep per mesh.
    """

    def __init__(self, quantity: Quantity, times: np.ndarray, n_t: int):
        self.quantity = quantity
        self.times = times
        self.n_t = n_t

        self.lower_bounds: Dict[int, np.ndarray] = dict()
        self.upper_bounds: Dict[int, np.ndarray] = dict()

        self.file_paths_be: Dict[int, str] = dict()
        self.file_paths_gbf: Dict[int, str] = dict()

    def _add_data(self, mesh: int, file_path_be: str, file_path_gbf: str, lower_bounds: np.ndarray,
                  upper_bounds: np.ndarray):
        self.file_paths_be[mesh] = file_path_be
        self.file_paths_gbf[mesh] = file_path_gbf

        self.lower_bounds[mesh] = lower_bounds
        self.upper_bounds[mesh] = upper_bounds

    def _load_data(self):
        self._vertices: Dict[int, np.ndarray] = dict()
        self._faces: Dict[int, np.ndarray] = dict()
        self._data: Dict[int, np.ndarray] = dict()

        for mesh in self.file_paths_be.keys():
            file_path_be = self.file_paths_be[mesh]
            file_path_gbf = self.file_paths_gbf[mesh]
            # Load .gbf
            with open(file_path_gbf, 'rb') as infile:
                offset = fdtype.INT.itemsize * 2 + fdtype.new(
                    (('i', 3),)).itemsize + fdtype.FLOAT.itemsize
                infile.seek(offset)

                dtype_meta = fdtype.new((('i', 3),))
                n_vertices, n_faces, _ = np.fromfile(infile, dtype_meta, 1)[0][1]

                dtype_vertices = fdtype.new((('f', 3 * n_vertices),))
                vertices = np.fromfile(infile, dtype_vertices, 1)[0][1].reshape(
                    (n_vertices, 3)).astype(float)

                dtype_faces = fdtype.new((('i', 3 * n_faces),))
                faces = fdtype.read(infile, dtype_faces, 1)[0][0].reshape((n_faces, 3)).astype(
                    int) - 1

            # Load .be
            data = np.empty((self.n_t,), dtype=object)
            with open(file_path_be, 'rb') as infile:
                offset = fdtype.INT.itemsize * 2
                infile.seek(offset)

                for t in range(self.n_t):
                    # Skip time value
                    infile.seek(fdtype.FLOAT.itemsize, 1)

                    dtype_meta = fdtype.new((('i', 4),))
                    n_faces = fdtype.read(infile, dtype_meta, 1)[0][0][3]
                    if n_faces > 0:
                        dtype_faces = fdtype.new((('f', n_faces),))
                        data[t] = np.fromfile(infile, dtype_faces, 1)[0][1]

            self._vertices[mesh] = vertices
            self._faces[mesh] = faces
            self._data[mesh] = data

    @property
    def vertices(self) -> Iterable:
        """Returns a global array of the vertices from all meshes.
        """
        if not hasattr(self, "_vertices"):
            self._load_data()

        size = sum(v.shape[0] for v in self._vertices.values())
        ret = np.empty((size, 3), dtype=float)
        counter = 0

        for v in self._vertices.values():
            size = v.shape[0]
            ret[counter:counter + size, :] = v
            counter += size

        return ret

    @property
    def faces(self) -> np.ndarray:
        """Returns a global array of the faces from all meshes.
        """
        if not hasattr(self, "_faces"):
            self._load_data()

        size = sum(f.shape[0] for f in self._faces.values())
        ret = np.empty((size, 3), dtype=int)
        counter = 0
        verts_counter = 0

        for m, f in self._faces.items():
            size = f.shape[0]
            ret[counter:counter + size, :] = f + verts_counter
            counter += size
            verts_counter += self._vertices[m].shape[0]

        return ret

    @property
    def data(self) -> np.ndarray:
        """Returns a global array of the loaded data for the quantity with data from all meshes.
        """
        if not hasattr(self, "_data"):
            self._load_data()

        ret = np.empty((self.n_t,), dtype=object)
        for t in range(self.n_t):
            size = sum(d[t].shape[0] for d in self._data.values())
            counter = 0
            ret[t] = np.empty((size,), dtype=float)

            for d in self._data.values():
                size = d[t].shape[0]
                ret[t][counter:counter + size] = d[t]
                counter += size

        return ret

    @property
    def vmin(self) -> float:
        """Minimum value of all faces at any time.
        """
        curr_min = min(np.min(b) for b in self.lower_bounds.values())
        if curr_min == 0.0:
            return min(min(np.min(p.data) for p in ps) for ps in self._faces.values())
        return curr_min

    @property
    def vmax(self) -> float:
        """Maximum value of all faces at any time.
        """
        curr_max = max(np.max(b) for b in self.upper_bounds.values())
        if curr_max == np.float32(-1e33):
            return max(max(np.max(p.data) for p in ps) for ps in self._faces.values())
        return curr_max


class Geometry:
    """Obstruction defined as a complex geometry.

    :ivar file_path: Path to the .ge file that defines the geom.
    :ivar texture_map: Path to the texture map of the geometry.
    :ivar texture_origin: Origin position of the texture provided by the surface.
    :ivar is_terrain: Indicates whether the geometry is a regular complex geom or terrain geometry.
    :ivar rgb: Color of the geometry in form of a 3-element tuple.
    :ivar surface: Surface object used for the geometry.
    """

    def __init__(self, file_path: str, texture_mapping: str,
                 texture_origin: Tuple[float, float, float],
                 is_terrain: bool, rgb: Tuple[int, int, int], surface: Surface = None):
        self.file_path = file_path
        self.texture_map = texture_mapping
        self.texture_origin = texture_origin
        self.is_terrain = is_terrain
        self.rgb = rgb
        self.surface = surface



import logging
import os
from copy import deepcopy
from typing import Dict, Sequence
import numpy as np

from fdsreader.utils import Quantity, Mesh
from fdsreader import settings
import fdsreader.utils.fortran_data as fdtype

_HANDLED_FUNCTIONS = {np.mean: (lambda pl: pl.mean)}


def implements(np_function):
    """Decorator to register an __array_function__ implementation for Slices.
    """

    def decorator(func):
        _HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


class SubPlot3D:
    """Subplot of a pl3d output for a single mesh.

    :ivar mesh: The mesh containing the data.
    """
    # Offset of the binary file to the end of the file header.
    _offset = fdtype.new((('i', 3),)).itemsize + fdtype.new((('i', 4),)).itemsize

    def __init__(self, file_path: str, mesh: Mesh):
        self.file_path = file_path  # Path to the binary data file
        self.mesh = mesh

    @property
    def data(self) -> np.ndarray:
        """Method to lazy load the 3D data for each quantity of a single mesh.

        :returns: 4D numpy array wiht (x,y,z,q) as dimensions, while q represents the 5 quantities.
        """
        if not hasattr(self, "_data"):
            with open(self.file_path, 'rb') as infile:
                dtype_data = fdtype.new((('f', self.mesh.dimension.size() * 5),))
                infile.seek(self._offset)
                self._data = fdtype.read(infile, dtype_data, 1)[0][0].reshape(
                    self.mesh.dimension.shape() + (5,), order='F')
        return self._data

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data


class Plot3D(np.lib.mixins.NDArrayOperatorsMixin):
    """Plot3d file data container including metadata. Consists of multiple subplots, one for each
        mesh.

    :ivar root_path: Path to the directory containing all slice files.
    :ivar time: The point in time when this data has been recorded.
    :ivar quantities: List with quantity objects containing information about recorded quantities
     calculated for this Plot3D with the corresponding label and unit.
    """

    def __init__(self, root_path: str, time: float, quantities: Sequence[Quantity]):
        self.root_path = root_path
        self.time = time
        self.quantities = quantities

        # List of all subplots this Plot3D consists of (one per mesh).
        self._subplots: Dict[Mesh, SubPlot3D] = dict()

    def _add_subplot(self, filename: str, mesh: Mesh) -> SubPlot3D:
        subplot = SubPlot3D(os.path.join(self.root_path, filename), mesh)
        self._subplots[mesh] = subplot

        # If lazy loading has been disabled by the user, load the data instantaneously instead
        if not settings.LAZY_LOAD:
            _ = subplot.data

        return subplot

    def __getitem__(self, mesh: Mesh):
        """Returns the :class:`SubPlot` that contains data for the given mesh.
        """
        return self._subplots[mesh]

    def mean(self) -> np.ndarray:
        """Calculates the mean over each quantity individually of the whole Plot3D.

        :returns: The calculated mean values.
        """
        mean_sums = np.zeros((5,))
        for subplot in self._subplots.values():
            mean_sums += np.mean(subplot.data, axis=(0, 1, 2))
        return mean_sums / len(self._subplots)

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for subplot in self._subplots.values():
            subplot.clear_cache()

    def __array__(self):
        """Method that will be called by numpy when trying to convert the object to a numpy ndarray.
        """
        raise UserWarning(
            "Plot3Ds can not be converted to numpy arrays, but they support all typical numpy"
            " operations such as np.multiply. If a 'global' array containg all subplots is"
            " required, please request this functionality by submitting an issue on Github.")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Method that will be called by numpy when using a ufunction with a Plot3D as input.

        :returns: A new pl3d on which the ufunc has been applied.
        """
        if method != "__call__":
            logging.warning(
                "The %s method has been used which is not explicitly implemented. Correctness of"
                " results is not guaranteed. If you require this feature to be implemented please"
                " submit an issue on Github where you explain your use case.", method)
        input_list = list(inputs)
        for i, inp in enumerate(inputs):
            if isinstance(inp, self.__class__):
                del input_list[i]

        new_pl3d = deepcopy(self)
        for subplot in self._subplots.values():
            subplot._data = ufunc(subplot.data, input_list[0], **kwargs)
        return new_pl3d

    def __array_function__(self, func, types, args, kwargs):
        """Method that will be called by numpy when using an array function with a Slice as input.

        :returns: The output of the array function.
        """
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
            # Note: this allows subclasses that don't override __array_function__ to handle Plot3Ds.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

# __array_function__ implementations
# ...



from typing import Iterable, Union, List, Tuple
import numpy as np

from fdsreader.bndf import Boundary, Obstruction
from fdsreader.utils.data import FDSDataCollection, Quantity


class ObstructionCollection(FDSDataCollection):
    """Collection of :class:`Obstruction` objects. Offers extensive functionality for filtering and
        using obstructions as well as dependent such as :class:`Boundary`.
    """

    def __init__(self, *obstructions: Iterable[Obstruction]):
        super().__init__(*obstructions)

    def filter_by_quantity(self, quantity: Union[str, Quantity]) -> List[Boundary]:
        """Filters all obstructions for its boundary data by a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        return [x.get_boundary_data(quantity) for x in self if
                x.get_boundary_data(quantity) is not None]

    def filter_by_boundary_data(self):
        return ObstructionCollection(x for x in self if x.has_boundary_data)

    def get_nearest_obstruction(self, point: Tuple[float, float, float]) -> Obstruction:
        """Filters the obstruction with the shortest distance to the given point.
        """
        d_min = np.finfo(float).min
        obst_min = None

        for obst in self:
            dx = max(obst.extent.x_start - point[0], 0, point[0] - obst.extent.x_end)
            dy = max(obst.extent.y_start - point[1], 0, point[1] - obst.extent.y_end)
            dz = max(obst.extent.z_start - point[2], 0, point[2] - obst.extent.z_end)
            d = np.sqrt(dx * dx + dy * dy + dz * dz)
            if d < d_min:
                d_min = d
                obst_min = obst

        return obst_min

    def get_border_obstructions(self):
        """Filters all obstructions with at least one face on the border of a mesh.
        """
        raise NotImplementedError("If you need this feature, please open an issue on GitHub "
                                  "describing your exact needs and your specific use case.")

    def __repr__(self):
        return "ObstructionCollection(" + super(ObstructionCollection, self).__repr__() + ")"



from typing import Union, Tuple, Dict

from fdsreader.utils import Surface, Extent, Mesh


class SubVentilation:
    """Part of a :class:`Ventilation`.

    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar mesh: The mesh that contains this part of the ventilation.
    """

    def __init__(self, mesh: Mesh, extent: Extent):
        self.mesh = mesh
        self.extent = extent


class Ventilation:
    """A ventilation can be used to model components of the ventilation system in a building, like a
        diffuser or a return.
        A ventilation can also be used as a means of applying a particular boundary condition to a
        rectangular patch on a solid surface.

    :ivar id: ID of the ventilation.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar surface: Surface object used for the ventilation.
    :ivar bound_indices: Indices used to define ventilation bounds in terms of mesh locations.
    :ivar color_index: Type of coloring used to color ventilation.
     -99 or +99 - use default color
     -n or +n - use n’th palette color
     < 0 - do not draw boundary file over vent
     > 0 - draw boundary file over vent
    :ivar draw_type: Defines how the ventilation is drawn.
     0 - solid surface
     2 - outline
     -2 - hidden
    :ivar open_time: Point in time the ventilation should open.
    :ivar close_time: Point in time the ventilation should close.
    :ivar rgba: Color of the ventilation in form of a 4-element tuple (ranging from 0.0 to 1.0).
    :ivar texture_origin: Origin position of the texture provided by the surface. When the texture does have a pattern,
     for example windows or bricks, the texture_origin specifies where the pattern should begin.
    :ivar circular_vent_origin: Origin of the ventilation relative to bounding box.
    :ivar radius: Radius of the ventilation circle.
    """

    def __init__(self, surface: Surface, bound_indices: Tuple[int, int, int, int, int, int],
                 color_index: int, draw_type: int,
                 rgba: Union[Tuple[()], Tuple[float, float, float, float]] = (),
                 texture_origin: Union[Tuple[()], Tuple[float, float, float]] = (),
                 circular_vent_origin: Union[Tuple[()], Tuple[float, float, float]] = (),
                 radius: float = -1):
        self.surface = surface
        self.bound_indices = bound_indices
        self.color_index = color_index
        self.draw_type = draw_type
        self.open_time = -1.0
        self.close_time = -1.0

        if len(rgba) != 0:
            self.rgba = rgba
        if len(texture_origin) != 0:
            self.texture_origin = texture_origin
        if len(circular_vent_origin) != 0:
            self.circular_vent_origin = circular_vent_origin
        if radius != -1:
            self.radius = radius

        self._subventilations: Dict[Mesh, SubVentilation] = dict()

    def _add_subventilation(self, mesh: Mesh, extent: Extent):
        self._subventilations[mesh] = SubVentilation(mesh, extent)

    def __repr__(self, *args, **kwargs):
        return f"Ventilation()"



LAZY_LOAD = True

ENABLE_CACHING = True

DEBUG = False

# As the binary representation of raw data is compiler dependent, this information must be provided
# by the user
FORTRAN_DATA_TYPE_INTEGER = "<i4"  # <i4 -> 32 bit integer (little-endian)
FORTRAN_DATA_TYPE_FLOAT = "<f4"  # <f4 -> 32 bit floating point (little-endian)
FORTRAN_DATA_TYPE_CHAR = "S"  # S -> 1-byte char
FORTRAN_BACKWARD = True  # Sets weather the blocks are ended with the size of the block



import logging
from functools import wraps
import sys
from fdsreader import settings


def log_error(module):
    def decorated(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                if settings.DEBUG:
                    raise e
                else:
                    e = type(e)(f"Module {str(module)}: {str(e)}\nThe error can be safely ignored if not requiring the {str(module)} module. However, please consider to submit an issue on Github including the error message, the stack trace and your FDS input-file so we can reproduce the error and fix it as soon as possible!").with_traceback(sys.exc_info()[2])
                    logging.exception(e)
        return wrapped
    return decorated



import os
from operator import attrgetter
from typing import List, TextIO, Dict, AnyStr, Sequence, Tuple, Union

import numpy as np
import pickle

from fdsreader.bndf import Obstruction, Patch, ObstructionCollection
from fdsreader.geom import Geometry, GeomBoundary, GeometryCollection
from fdsreader.isof import Isosurface, IsosurfaceCollection
from fdsreader.part import Particle, ParticleCollection
from fdsreader.pl3d import Plot3D, Plot3DCollection
from fdsreader.slcf import Slice, SliceCollection
from fdsreader.utils import Mesh, Dimension, Surface, Quantity, Ventilation, Extent, log_error
from fdsreader.utils.data import create_hash, get_smv_file, Device
import fdsreader.utils.fortran_data as fdtype
from fdsreader import settings
from fdsreader._version import __version__


class Simulation:
    """Master class managing all data for a given simulation.

    :ivar smv_file_path: Path to the .smv file of the simulation.
    :ivar root_path: Path to the root directory of the simulation.
    :ivar fds_version: Version of FDS the simulation was performed with.
    :ivar chid: Name (ID) of the simulation.
    :ivar hrrpuv_cutoff: The hrrpuv_cutoff value.
    :ivar default_texture_origin: The default origin used for textures with no explicit origin.
    :ivar out_file_path: Path to the .out file of the simulation.
    :ivar surfaces: List containing all surfaces defined in this simulation.
    :ivar meshes: List containing all meshes (grids) defined in this simulation.
    :ivar ventilations: List containing all ventilations defined in this simulation.
    :ivar obstructions: All defined obstructions combined into a :class:`ObstructionCollection`.
    :ivar slices: All defined slices combined into a :class:`SliceCollection`.
    :ivar data_3d: All defined 3D plotting data combined into a :class:`Plot3DCollection`.
    :ivar isosurfaces: All defined isosurfaces combined into a :class:`IsosurfaceCollection`.
    :ivar particles: All defined particles combined into a :class:`ParticleCollection`.
    :ivar devices: List containing all devices defined in this simulation.
    :ivar geoms: List containing all geometries defined in this simulation.
    :ivar geom_data: All geometry data by quantity combined into a :class:`GeometryCollection`.
    :ivar cpu: Dictionary mapping .csv header keys to numpy arrays containing cpu data.
    :ivar hrr: Dictionary mapping .csv header keys to numpy arrays containing hrr data.
    :ivar steps: Dictionary mapping .csv header keys to numpy arrays containing steps data.
    """

    def __new__(cls, path: str):
        if settings.ENABLE_CACHING:
            smv_file_path = get_smv_file(path)
            root_path = os.path.dirname(path)

            with open(smv_file_path, 'r') as infile:
                for line in infile:
                    if line.strip() == "CHID":
                        chid = infile.readline().strip()
                        break

            pickle_file_path = Simulation._get_pickle_filename(root_path, chid)

            if os.path.isfile(pickle_file_path):
                try:
                    with open(pickle_file_path) as f:
                        sim = pickle.load(f)
                    # Check if pickle file is valid
                    if isinstance(sim, cls) and sim.reader_version == __version__ and sim._hash == create_hash(smv_file_path):
                        return sim
                except:
                    os.remove(pickle_file_path)

        return super(Simulation, cls).__new__(cls)

    def __init__(self, path: str):
        """
        :param path: Either the path to the directory containing the simulation data or direct path
            to the .smv file for the simulation in case that multiple simulation output was written to
            the same directory.
        """
        # Check if the file has already been instantiated via a cached pickle file
        if not hasattr(self, "_hash"):
            self.reader_version = __version__

            self.smv_file_path = get_smv_file(path)

            self.root_path = os.path.dirname(self.smv_file_path)

            self.geoms: List[Geometry] = list()
            self.meshes: List[Mesh] = list()
            self.surfaces: List[Surface] = list()
            self.obstructions = dict()
            self.ventilations = dict()

            # First collect all meta-information for any FDS data to later combine the gathered
            # information into data collections
            self.slices = dict()
            self.data_3d = dict()
            self.isosurfaces = dict()
            self.particles = list()
            self.geom_data = list()
            self.devices: Dict[str, Union[Device, List[Device]]] = dict()
            device_tmp = str()

            with open(self.smv_file_path, 'r') as smv_file:
                for line in smv_file:
                    keyword = line.strip()
                    if keyword == "VERSION":
                        self.fds_version = smv_file.readline().strip()
                    elif keyword == "CHID":
                        self.chid = smv_file.readline().strip()
                    elif keyword == "CSVF":
                        csv_type = smv_file.readline().strip()
                        filename = smv_file.readline().strip()
                        file_path = os.path.join(self.root_path, filename)
                        if csv_type == "hrr":
                            self.hrr = self._load_HRR_data(file_path)
                        elif csv_type == "steps":
                            self.steps = self._load_step_data(file_path)
                        elif csv_type == "devc":
                            device_tmp = file_path
                            self.devices["Time"] = Device(Quantity("Time", "Time", ""),
                                                          (.0, .0, .0), (.0, .0, .0))
                    elif keyword == "HRRPUVCUT":
                        self.hrrpuv_cutoff = float(smv_file.readline().strip())
                    elif keyword == "TOFFSET":
                        offsets = smv_file.readline().strip().split()
                        self.default_texture_origin = tuple(float(offsets[i]) for i in range(3))
                    elif keyword == "CLASS_OF_PARTICLES":
                        self.particles.append(self._register_particle(smv_file))
                    elif keyword.startswith("GEOM"):
                        self._load_geoms(smv_file, keyword)
                    elif keyword.startswith("GRID"):
                        self.meshes.append(self._load_mesh(smv_file, keyword))
                    elif keyword == "SURFACE":
                        self.surfaces.append(self._load_surface(smv_file))
                    elif keyword == "DEVICE":
                        name, device = self._register_device(smv_file)
                        if name in self.devices:
                            if type(self.devices) == list:
                                self.devices[name].append(device)
                            else:
                                self.devices[name] = [self.devices[name], device]
                        else:
                            self.devices[name] = device
                    elif keyword.startswith("SLC"):
                        self._load_slice(smv_file, keyword)
                    elif "ISOG" in keyword:
                        self._load_isosurface(smv_file, keyword)
                    elif keyword.startswith("PL3D"):
                        self._load_data_3d(smv_file, keyword)
                    elif keyword.startswith("BNDF"):
                        self._load_boundary_data(smv_file, keyword, cell_centered=False)
                    elif keyword.startswith("BNDC"):
                        self._load_boundary_data(smv_file, keyword, cell_centered=True)
                    elif keyword.startswith("BNDE"):
                        self._load_boundary_data_geom(smv_file, keyword)
                    elif keyword.startswith("PRT5"):
                        self._load_particle_data(smv_file, keyword)

                self.cpu = self._load_CPU_data()
                if device_tmp != "":
                    self._load_DEVC_data(device_tmp)

            # POST INIT (post read)
            self.out_file_path = os.path.join(self.root_path, self.chid + ".out")
            self.ventilations: List[Ventilation] = list(self.ventilations.values())
            self.obstructions = ObstructionCollection(self._merged_obstructions)
            for obst in self.obstructions:
                obst._post_init()
            # Combine the gathered temporary information into data collections
            self.geom_data = GeometryCollection(self.geom_data)
            self.slices = SliceCollection(
                Slice(self.root_path, slice_data[0]["id"], slice_data[0]["cell_centered"],
                      slice_data[0]["times"], slice_data[1:]) for slice_data in
                self.slices.values())
            self.data_3d = Plot3DCollection(self.data_3d.keys(), self.data_3d.values())
            self.isosurfaces = IsosurfaceCollection(self.isosurfaces.values())
            if self.particles is None:
                self.particles = ParticleCollection(())

            if settings.ENABLE_CACHING:
                # Hash will be saved to simulation pickle file and compared to new hash when loading
                # the pickled simulation again in the next run of the prpogram.
                self._hash = create_hash(self.smv_file_path)
                pickle.dump(self,
                            open(Simulation._get_pickle_filename(self.root_path, self.chid), 'wb'),
                            protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def _get_pickle_filename(cls, root_path: str, chid: str) -> AnyStr:
        """Get the filename used to save the pickled simulation.
        """
        return os.path.join(root_path, chid + ".pickle")

    def _load_mesh(self, smv_file: TextIO, line: str) -> Mesh:
        """Load information for a single mesh from the smv file at current pointer position.
        """
        mesh_id = line.split()[1]

        grid_numbers = smv_file.readline().strip().split()
        grid_dimensions = {'x': int(grid_numbers[0]) + 1, 'y': int(grid_numbers[1]) + 1,
                           'z': int(grid_numbers[2]) + 1}

        smv_file.readline()  # Blank line
        assert smv_file.readline().strip() == "PDIM"
        coordinates = dict()
        extents = smv_file.readline().split()
        extents = {'x': (float(extents[0]), float(extents[1])),
                   'y': (float(extents[2]), float(extents[3])),
                   'z': (float(extents[4]), float(extents[5]))}

        for dim in ('x', 'y', 'z'):
            smv_file.readline()  # Blank line
            assert smv_file.readline().strip()[:3] == "TRN"
            noc = int(smv_file.readline().strip())
            for _ in range(noc):
                smv_file.readline()
            coordinates[dim] = np.empty(grid_dimensions[dim], dtype=np.float32)
            for i in range(grid_dimensions[dim]):
                coordinates[dim][i] = float(smv_file.readline().split()[1])

        mesh = Mesh(coordinates, extents, mesh_id)

        smv_file.readline()  # Blank line
        assert smv_file.readline().strip() == "OBST"
        self._load_obstructions(smv_file, mesh)

        smv_file.readline()  # Blank line
        assert smv_file.readline().strip() == "VENT"
        self._load_vents(smv_file, mesh)

        return mesh

    @log_error("obst")
    def _load_obstructions(self, smv_file: TextIO, mesh: Mesh):
        temp_data = list()
        n = int(smv_file.readline().strip())

        if n != 0:
            self.obstructions[mesh] = list()

        for _ in range(n):
            line = smv_file.readline().strip().split()
            ext = [float(line[i]) for i in range(6)]
            obst_ordinal = int(line[6])

            side_surfaces = tuple(self.surfaces[int(line[i]) - 1] for i in range(7, 13))
            if len(line) > 13:
                texture_origin = (float(line[13]), float(line[14]), float(line[15]))
                temp_data.append((obst_ordinal, Extent(*ext), side_surfaces, texture_origin))
            else:
                temp_data.append((obst_ordinal, Extent(*ext), side_surfaces))

        # We need to save the extents of obstructions that have been created due to a hole to later
        # recreate a Hole with the correct extents
        hole_extents: Dict[int, List[Extent]] = dict()

        for tmp in temp_data:
            line = smv_file.readline().strip().split()
            bound_indices = (
                int(float(line[0])) - 1, int(float(line[1])) - 1, int(float(line[2])) - 1,
                int(float(line[3])) - 1, int(float(line[4])) - 1, int(float(line[5])) - 1)
            color_index = int(line[6])
            block_type = int(line[7])
            if color_index == -3:
                rgba = tuple(float(line[i]) for i in range(8, 12))
            else:
                rgba = ()

            if len(tmp) == 4:
                obst_ordinal, extent, side_surfaces, texture_origin = tmp
            else:
                obst_ordinal, extent, side_surfaces = tmp
                texture_origin = self.default_texture_origin

            obst_tmp = Obstruction(obst_ordinal, side_surfaces, bound_indices, color_index,
                                   block_type, texture_origin, rgba=rgba)
            obst = next((o for o in self._merged_obstructions if o.id == obst_tmp.id), obst_tmp)

            self.obstructions[mesh].append(obst)
            # Holes will be handles differently
            if obst_ordinal < 0:
                if mesh not in obst._extents:
                    obst._extents[mesh] = list()
                obst._extents[mesh].append(extent)
            else:
                obst._extents[mesh] = extent
                mesh.obstructions.append(obst)

            if obst_ordinal < 0:
                if obst_ordinal not in hole_extents:
                    hole_extents[obst_ordinal] = list()
                hole_extents[obst_ordinal].append(extent)

        for ordinal, extents in hole_extents.items():
            hole_extent = list()

            minx = min(extents, key=attrgetter('x_start'))
            hole_extent.append(minx.x_end)
            maxx = max(extents, key=attrgetter('x_end'))
            hole_extent.append(maxx.x_start)
            miny = min(extents, key=attrgetter('y_start'))
            hole_extent.append(miny.y_end)
            maxy = max(extents, key=attrgetter('y_end'))
            hole_extent.append(maxy.y_start)
            minz = min(extents, key=attrgetter('z_start'))
            hole_extent.append(minz.z_end)
            maxz = max(extents, key=attrgetter('z_end'))
            hole_extent.append(maxz.z_start)

            hole_extent = Extent(*hole_extent)
            obst = next(o for o in self._merged_obstructions if o.id == ordinal)
            obst._extents[mesh].append(hole_extent)

    def _load_geoms(self, smv_file: TextIO, line: str):
        ngeoms = int(line.split()[1])

        filename = smv_file.readline()
        file_path = os.path.join(self.root_path, filename)

        for g in range(ngeoms):
            line = smv_file.readline().split("!")
            texture_line = line[0].split()
            rgb_line = line[1].split()

            texture_mapping = texture_line[0]
            texture_origin = (
                float(texture_line[1]), float(texture_line[2]), float(texture_line[3]))
            is_terrain = bool(texture_line[4])
            rgb = (int(rgb_line[0]), int(rgb_line[1]), int(rgb_line[2]))

            if '%' in line[0]:
                surface_id = line[0].split("%")[-1]
                surface = next((s for s in self.surfaces if s.id() == surface_id), None)
                geom = Geometry(file_path, texture_mapping, texture_origin, is_terrain, rgb,
                                surface=surface)
            else:
                geom = Geometry(file_path, texture_mapping, texture_origin, is_terrain, rgb)
            self.geoms.append(geom)

    @property
    def _merged_obstructions(self) -> List[Obstruction]:
        ret = list()
        for _, obstructions in self.obstructions.items():
            for obst in obstructions:
                if not any(obst.id == o.id for o in ret):
                    ret.append(obst)
        return ret

    @log_error("vents")
    def _load_vents(self, smv_file: TextIO, mesh: Mesh):
        line = smv_file.readline().split()
        n, n_dummies = int(line[0]), int(line[1])

        temp_data = list()

        def read_common_info():
            line = smv_file.readline().strip().split()
            return line, [float(line[i]) for i in range(6)], int(line[6]) - 1, self.surfaces[
                int(line[7])]

        def read_common_info2():
            line = smv_file.readline().strip().split()
            bound_indices = tuple(int(line[i]) for i in range(6))
            color_index = int(line[6])
            draw_type = int(line[7])
            if len(line) > 8:
                rgba = tuple(float(line[i]) for i in range(8, 12))
            else:
                rgba = ()
            return bound_indices, color_index, draw_type, rgba

        texture_origin = ()
        for _ in range(n - n_dummies):
            line, ext, vent_index, surface = read_common_info()
            texture_origin = (float(line[8]), float(line[9]), float(line[10]))
            temp_data.append((Extent(*ext), vent_index, surface, texture_origin))

        for _ in range(n_dummies):
            _, ext, vent_index, surface = read_common_info()
            temp_data.append((Extent(*ext), vent_index, surface))

        for v in range(n):
            if v < n - n_dummies:
                extent, vent_index, surface, texture_origin = temp_data[v]
            else:
                extent, vent_index, surface = temp_data[v]
            bound_indices, color_index, draw_type, rgba = read_common_info2()
            if vent_index not in self.ventilations:
                self.ventilations[vent_index] = Ventilation(surface, bound_indices, color_index,
                                                            draw_type, rgba=rgba,
                                                            texture_origin=texture_origin)
            self.ventilations[vent_index]._add_subventilation(mesh, extent)

        smv_file.readline()
        assert "CVENT" in smv_file.readline()

        n = int(smv_file.readline().strip())
        temp_data.clear()
        for _ in range(n):
            line, extent, vent_index, surface = read_common_info()
            circular_vent_origin = (float(line[12]), float(line[13]), float(line[14]))
            radius = float(line[15])
            temp_data.append(
                (extent, vent_index, surface, texture_origin, circular_vent_origin, radius))

        for v in range(n):
            extent, vent_index, surface, texture_origin, circular_vent_origin, radius = temp_data[v]
            bound_indices, color_index, draw_type, rgba = read_common_info2()
            if vent_index not in self.ventilations:
                self.ventilations[vent_index] = Ventilation(surface, bound_indices,
                                                            color_index, draw_type, rgba=rgba,
                                                            texture_origin=texture_origin,
                                                            circular_vent_origin=circular_vent_origin,
                                                            radius=radius)
            self.ventilations[vent_index]._add_subventilation(mesh, extent)

    @log_error("surface")
    def _load_surface(self, smv_file: TextIO) -> Surface:
        """Load the information for a single surface from the smv file at current pointer position.
        """

        surface_id = smv_file.readline().strip()

        line = smv_file.readline().strip().split()
        tmpm, material_emissivity = float(line[0]), float(line[1])

        line = smv_file.readline().strip().split()

        surface_type = int(line[0])
        texture_width, texture_height = float(line[1]), float(line[2])
        rgb = (float(line[3]), float(line[4]), float(line[5]))
        transparency = float(line[6])

        texture_map = smv_file.readline().strip()
        texture_map = None if texture_map == "null" else os.path.join(self.root_path, texture_map)

        return Surface(surface_id, tmpm, material_emissivity, surface_type, texture_width,
                       texture_height, texture_map, rgb, transparency)

    @log_error("slcf")
    def _load_slice(self, smv_file: TextIO, line: str):
        """Loads the slice at current pointer position.
        """
        if "SLCC" in line:
            cell_centered = True
        else:
            cell_centered = False

        slice_index = int(line.split('!')[1].strip().split()[0])

        slice_id = line.split('%')[1].split('&')[0].strip() if '%' in line else ""

        mesh_index = int(line.split('&')[0].strip().split()[1]) - 1
        mesh = self.meshes[mesh_index]

        # Read in index ranges for x, y and z
        bound_indices = [int(i.strip()) for i in line.split('&')[1].split('!')[0].strip().split()]
        extent, dimension = self._indices_to_extent(bound_indices, mesh)

        filename = smv_file.readline().strip()
        quantity = smv_file.readline().strip()
        label = smv_file.readline().strip()
        unit = smv_file.readline().strip()

        file_path = os.path.join(self.root_path, filename)

        if os.path.exists(file_path + ".bnd"):
            times = list()
            with open(file_path + ".bnd", 'r') as bnd_file:
                for line in bnd_file:
                    times.append(float(line.split()[0]))
            times = np.array(times)
        else:
            times = None

        if slice_index not in self.slices:
            self.slices[slice_index] = [
                {"cell_centered": cell_centered, "times": times, "id": slice_id}]
        self.slices[slice_index].append(
            {"dimension": dimension, "extent": extent, "mesh": mesh, "filename": filename,
             "quantity": quantity, "label": label, "unit": unit})

    @log_error("bndf")
    def _load_boundary_data(self, smv_file: TextIO, line: str, cell_centered: bool):
        """Loads the boundary data at current pointer position.
        """
        line = line.split()
        mesh_index = int(line[1]) - 1
        mesh = self.meshes[mesh_index]

        filename = smv_file.readline().strip()
        quantity = smv_file.readline().strip()
        label = smv_file.readline().strip()
        unit = smv_file.readline().strip()

        bid = int(filename.split('_')[-1][:-3])

        file_path = os.path.join(self.root_path, filename)

        patches = dict()

        if os.path.exists(file_path + ".bnd"):
            times = list()
            lower_bounds = list()
            upper_bounds = list()
            with open(file_path + ".bnd", 'r') as bnd_file:
                for line in bnd_file:
                    splits = line.split()
                    times.append(float(splits[0]))
                    lower_bounds.append(float(splits[1]))
                    upper_bounds.append(float(splits[2]))
            times = np.array(times)
            lower_bounds = np.array(lower_bounds, dtype=np.float32)
            upper_bounds = np.array(upper_bounds, dtype=np.float32)
            n_t = times.shape[0]
        else:
            times = None
            n_t = -1
            lower_bounds = np.array([0.0], dtype=np.float32)
            upper_bounds = np.array([np.float32(-1e33)], dtype=np.float32)

        with open(file_path, 'rb') as infile:
            # Offset of the binary file to the end of the file header.
            offset = 3 * fdtype.new((('c', 30),)).itemsize
            infile.seek(offset)

            n_patches = fdtype.read(infile, fdtype.INT, 1)[0][0][0]

            dtype_patches = fdtype.new((('i', 9),))
            patch_infos = fdtype.read(infile, dtype_patches, n_patches)
            offset += fdtype.INT.itemsize + dtype_patches.itemsize * n_patches
            patch_offset = fdtype.FLOAT.itemsize

            for patch_info in patch_infos:
                patch_info = patch_info[0]

                extent, dimension = self._indices_to_extent(patch_info[:6], mesh)
                orientation = patch_info[6]
                obst_index = patch_info[7]

                p = Patch(file_path, dimension, extent, orientation, cell_centered,
                          patch_offset, offset, n_t)

                # Skip obstacles with index 0, which just gives the extent of the (whole) mesh faces
                # These might be needed in case of "closed" mesh faces
                if obst_index != 0:
                    obst_index -= 1  # Account for fortran indexing
                    if obst_index not in patches:
                        patches[obst_index] = list()
                    patches[obst_index].append(p)
                patch_offset += fdtype.new(
                    (('f', str(p.dimension.shape(cell_centered=False))),)).itemsize

        for obst_index, p in patches.items():
            for patch in p:
                patch._post_init(patch_offset)

            self.obstructions[mesh][obst_index]._add_patches(bid, cell_centered, quantity, label,
                                                             unit, mesh, p, times, n_t,
                                                             lower_bounds, upper_bounds)

    @log_error("geom")
    def _load_boundary_data_geom(self, smv_file: TextIO, line: str):
        line = line.split()
        mesh_index = int(line[1]) - 1
        # Meshes are not loaded yet
        # mesh = self.meshes[mesh_index]

        filename_be = smv_file.readline().strip()
        filename_gbf = smv_file.readline().strip()
        quantity = smv_file.readline().strip()
        label = smv_file.readline().strip()
        unit = smv_file.readline().strip()

        bid = int(filename_be.split('_')[-1][:-3]) - 1

        file_path_be = os.path.join(self.root_path, filename_be)
        file_path_gbf = os.path.join(self.root_path, filename_gbf)

        times = list()
        lower_bounds = list()
        upper_bounds = list()
        with open(file_path_be + ".bnd", 'r') as bnd_file:
            for line in bnd_file:
                splits = line.split()
                times.append(float(splits[0]))
                lower_bounds.append(float(splits[1]))
                upper_bounds.append(float(splits[2]))
        times = np.array(times)
        lower_bounds = np.array(lower_bounds, dtype=np.float32)
        upper_bounds = np.array(upper_bounds, dtype=np.float32)
        n_t = times.shape[0]

        if bid >= len(self.geom_data):
            self.geom_data.append(GeomBoundary(Quantity(quantity, label, unit), times, n_t))
        self.geom_data[bid]._add_data(mesh_index, file_path_be, file_path_gbf, lower_bounds,
                                      upper_bounds)

    @log_error("pl3d")
    def _load_data_3d(self, smv_file: TextIO, line: str):
        """Loads the pl3d at current pointer position.
        """
        line = line.strip().split()

        time = float(line[1])

        mesh_index = int(line[2]) - 1

        filename = smv_file.readline().strip()
        quantities = list()
        for _ in range(5):
            quantity = smv_file.readline().strip()
            label = smv_file.readline().strip()
            unit = smv_file.readline().strip()
            quantities.append(Quantity(quantity, label, unit))

        if time not in self.data_3d:
            self.data_3d[time] = Plot3D(self.root_path, time, quantities)
        self.data_3d[time]._add_subplot(filename, self.meshes[mesh_index])

    @log_error("isof")
    def _load_isosurface(self, smv_file: TextIO, line: str):
        """Loads the isosurface at current pointer position.
        """
        double_quantity = line[0] == 'T'
        mesh_index = int(line.strip().split()[1]) - 1

        iso_filename = smv_file.readline().strip()
        iso_id = int(iso_filename.split('_')[-1][:-4])
        iso_file_path = os.path.join(self.root_path, iso_filename)

        if double_quantity:
            viso_file_path = os.path.join(self.root_path, smv_file.readline().strip())
        quantity = smv_file.readline().strip()
        label = smv_file.readline().strip()
        unit = smv_file.readline().strip()
        if double_quantity:
            v_quantity = smv_file.readline().strip()
            v_label = smv_file.readline().strip()
            v_unit = smv_file.readline().strip()

        if iso_id not in self.isosurfaces:
            with open(iso_file_path, 'rb') as infile:
                nlevels = fdtype.read(infile, fdtype.INT, 3)[2][0][0]

                dtype_header_levels = fdtype.new((('f', nlevels),))
                levels = fdtype.read(infile, dtype_header_levels, 1)[0]
        if double_quantity:
            if iso_id not in self.isosurfaces:
                self.isosurfaces[iso_id] = Isosurface(iso_id, double_quantity, quantity, label,
                                                      unit, levels, v_quantity=v_quantity,
                                                      v_label=v_label, v_unit=v_unit)
            self.isosurfaces[iso_id]._add_subsurface(self.meshes[mesh_index], iso_file_path,
                                                     viso_file_path=viso_file_path)
        else:
            if iso_id not in self.isosurfaces:
                self.isosurfaces[iso_id] = Isosurface(iso_id, double_quantity, quantity, label,
                                                      unit, levels)
            self.isosurfaces[iso_id]._add_subsurface(self.meshes[mesh_index], iso_file_path)

    @log_error("part")
    def _register_particle(self, smv_file: TextIO) -> Particle:
        particle_class = smv_file.readline().strip()
        color = tuple(float(c) for c in smv_file.readline().strip().split())

        n_q = int(smv_file.readline().strip())
        quantities = list()
        for _ in range(n_q):
            quantity = smv_file.readline().strip()
            label = smv_file.readline().strip()
            unit = smv_file.readline().strip()
            quantities.append(Quantity(quantity, label, unit))
        return Particle(particle_class, quantities, color)

    def _load_particle_meta(self, particles: Union[List[Particle], ParticleCollection],
                            file_path: str, mesh: Mesh) -> List[float]:
        with open(file_path, 'r') as bnd_file:
            line = bnd_file.readline().strip().split()
            n_classes = int(line[1])
            times = list()
            n_q = list()
            for i in range(n_classes):
                line = bnd_file.readline().strip().split()
                n_q.append(int(line[0]))
                for _ in range(n_q[-1]):
                    bnd_file.readline()
                particles[i].n_particles[mesh] = list()
            bnd_file.seek(0)

            for line in bnd_file:
                times.append(float(line.strip().split()[0]))
                for i in range(n_classes):
                    particle = particles[i]
                    particle.n_particles[mesh].append(
                        int(bnd_file.readline().strip().split()[1].strip()))
                    for q in range(n_q[i]):
                        line = bnd_file.readline().strip().split()
                        quantity = particle.quantities[q].quantity
                        particle.lower_bounds[quantity].append(float(line[0]))
                        particle.upper_bounds[quantity].append(float(line[1]))
        return times

    @log_error("part")
    def _load_particle_data(self, smv_file: TextIO, line: str):
        file_path = os.path.join(self.root_path, smv_file.readline().strip())

        mesh_index = int(line.split()[1].strip()) - 1
        mesh = self.meshes[mesh_index]

        times = self._load_particle_meta(self.particles, file_path + '.bnd', mesh)
        if type(self.particles) == list:
            self.particles = ParticleCollection(times, self.particles)

        self.particles._file_paths[mesh] = file_path

        n_classes = int(smv_file.readline().strip())
        for i in range(n_classes):
            smv_file.readline()

    @log_error("devc")
    def _register_device(self, smv_file: TextIO) -> Tuple[str, Device]:
        line = smv_file.readline().strip().split('%')
        name = line[0].strip()
        quantity_label = line[1].strip()
        line = smv_file.readline().strip().split('#')[0].split()
        position = (float(line[0]), float(line[1]), float(line[2]))
        orientation = (float(line[3]), float(line[4]), float(line[5]))
        return name, Device(Quantity(name, quantity_label, ""), position, orientation)

    @log_error("devc")
    def _load_DEVC_data(self, file_path: str):
        with open(file_path, 'r') as infile:
            units = infile.readline().split(',')
            names = [name.replace('"', '').replace('\n', '').strip() for name in
                     infile.readline().split(',')]
            values = np.genfromtxt(infile, delimiter=',', dtype=np.float32, autostrip=True)
            for k in range(len(names)):
                devc = self.devices[names[k]]
                devc.quantity.unit = units[k]
                size = values.shape[0]
                devc.data = np.empty((size,), dtype=np.float32)
                for i in range(size):
                    devc.data[i] = values[i][k]

        line_path = file_path.replace("devc", "steps")
        if os.path.exists(line_path):
            with open(line_path, 'r') as infile:
                infile.readline()
                names = infile.readline().split(',')
                data = np.genfromtxt(infile, delimiter=',', dtype=np.float32, autostrip=True)
                for k, key in enumerate(names):
                    if key in self.devices:
                        devc = self.devices[key]
                        for i in range(len(devc)):
                            devc[i] = data[k, i]

    @log_error("csv")
    def _load_HRR_data(self, file_path: str) -> Dict[str, np.ndarray]:
        with open(file_path, 'r') as infile:
            infile.readline()
            keys = infile.readline().split(',')
            values = np.genfromtxt(infile, delimiter=',', dtype=np.float32, autostrip=True)
            return self._transform_csv_data(keys, values, [np.float32] * len(keys))

    @log_error("csv")
    def _load_step_data(self, file_path: str) -> Dict[str, np.ndarray]:
        with open(file_path, 'r') as infile:
            infile.readline()
            keys = infile.readline().split(',')
            dtypes = [np.float32] * len(keys)
            dtypes[0] = int
            dtypes[1] = np.dtype("datetime64[ms]")
            values = np.genfromtxt(infile, delimiter=',', dtype=dtypes, autostrip=True)
            return self._transform_csv_data(keys, values, dtypes)

    @log_error("csv")
    def _load_CPU_data(self) -> Dict[str, np.ndarray]:
        file_path = os.path.join(self.root_path, self.chid + "_cpu.csv")
        if os.path.exists(file_path):
            with open(file_path, 'r') as infile:
                keys = infile.readline().split(",")
                dtypes = [np.float32] * len(keys)
                dtypes[0] = int
                values = np.genfromtxt(infile, delimiter=',', dtype=dtypes, autostrip=True)
                if len(values.shape) != 0:
                    return self._transform_csv_data(keys, values, dtypes)
                else:
                    return self._transform_csv_data(keys, values.reshape((1,)), dtypes)

    def _transform_csv_data(self, keys, values, dtypes):
        size = values.shape[0]
        data = {keys[i]: np.empty((size,), dtype=dtypes[i]) for i in range(len(keys))}
        for k, arr in enumerate(data.values()):
            for i in range(size):
                arr[i] = values[i][k]
        return data

    def _indices_to_extent(self, indices: Sequence[Union[int, str]], mesh: Mesh) -> Tuple[
        Extent, Dimension]:
        co = mesh.coordinates
        x_min, x_max, y_min, y_max, z_min, z_max = (
            int(indices[0]), int(indices[1]), int(indices[2]), int(indices[3]), int(indices[4]),
            int(indices[5]))
        co_x_min, co_x_max, co_y_min, co_y_max, co_z_min, co_z_max = (
            co['x'][x_min], co['x'][x_max], co['y'][y_min],
            co['y'][y_max], co['z'][z_min], co['z'][z_max])
        dimension = Dimension(x_max - x_min + 1, y_max - y_min + 1, z_max - z_min + 1)

        extent = Extent(co_x_min, co_x_max, co_y_min, co_y_max, co_z_min, co_z_max)
        return extent, dimension

    def clear_cache(self, clear_persistent_cache=False):
        """Remove all data from the internal cache that has been loaded so far to free memory.

        :param clear_persistent_cache: Whether to clear the persistent simulation cache as well.
        """
        self.slices.clear_cache()
        self.data_3d.clear_cache()
        self.isosurfaces.clear_cache()
        self.particles.clear_cache()
        self.obstructions.clear_cache()

        if clear_persistent_cache:
            os.remove(Simulation._get_pickle_filename(self.root_path, self.chid))



from .data import Quantity

from fdsreader.utils.dimension import Dimension

from fdsreader.utils.extent import Extent

from .fds_classes.surface import Surface

from .fds_classes.mesh import Mesh

from .fds_classes.ventilation import Ventilation

from.misc import *


from typing import Iterable, Union, List

from fdsreader.pl3d import Plot3D
from fdsreader.utils.data import FDSDataCollection, Quantity


class Plot3DCollection(FDSDataCollection):
    """Collection of :class:`Plot3D` objects. Offers extensive functionality for filtering and
        using plot3Ds as well as its subclasses such as :class:`SubPlot3D`.
    """

    def __init__(self, times: Iterable[float], *plot3ds: Iterable[Plot3D]):
        super().__init__(*plot3ds)
        self.times = list(times)

    def filter_by_quantity(self, quantity: Union[str, Quantity]):
        """Filters all plot3d data by a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        return Plot3DCollection(x for x in self if any(
            q.quantity.lower() == quantity.lower() or q.label.lower() == quantity.lower() for q in
            x.quantities))

    def __repr__(self):
        return "Plot3DCollection(" + super(Plot3DCollection, self).__repr__() + ")"



from typing import Dict, Tuple
from typing_extensions import Literal
import numpy as np

from fdsreader.utils import Dimension, Extent


class Mesh:
    """3-dimensional Mesh of fixed, defined size.

    :ivar coordinates: Coordinate values for each of the 3 dimension.
    :ivar dimensions: :class:`Dimension` describing the size of the 3 dimensions regarding indices.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar n: Number of elements for each of the 3 dimensions.
    :ivar n_size: Total number of blocks in this mesh.
    :ivar id: Mesh id/label assigned to this mesh.
    """
    id = None  # Needed for hash to work

    def __init__(self, coordinates: Dict[Literal['x', 'y', 'z'], np.ndarray],
                 extents: Dict[Literal['x', 'y', 'z'], Tuple[float, float]], mesh_id: str):
        """
        :param coordinates: Coordinate values of the three axes.
        :param extents: Extent of the mesh in each dimension.
        :param mesh_id: ID of this mesh.
        """
        self.id = mesh_id
        self.coordinates = coordinates
        self.dimension = Dimension(
            coordinates['x'].size if coordinates['x'].size > 0 else 1,
            coordinates['y'].size if coordinates['y'].size > 0 else 1,
            coordinates['z'].size if coordinates['z'].size > 0 else 1)

        self.n_size = self.dimension.size()
        self.extent = Extent(extents['x'][0], extents['x'][1], extents['y'][0], extents['y'][1],
                             extents['z'][0], extents['z'][1])

        self.obstructions = list()

    def get_obstruction_mask(self, cell_centered=False) -> np.ndarray:
        shape = self.dimension.shape(cell_centered=cell_centered)
        mask = np.ones(shape, dtype=bool)

        for obst in self.obstructions:
            x1, x2 = obst.bound_indices['x']
            y1, y2 = obst.bound_indices['y']
            z1, z2 = obst.bound_indices['z']
            c = 1 if cell_centered else 0
            mask[x1:max(x2 + c, x1 + 1), y1:max(y2 + c, y1 + 1), z1:max(z2 + c, z1 + 1)] = False

        return mask

    def get_obstruction_mask_slice(self, orientation: Literal[1, 2, 3], value: float,
                                   cell_centered=False):
        assert orientation in (1, 2, 3), "The slices orientation has to be one of [1,2,3]!"

        slc_index = self.coordinate_to_index((value,), dimension=(orientation - 1,),
                                             cell_centered=cell_centered)[0]

        mask_indices = [slice(None)] * 3
        mask_indices[orientation - 1] = slice(slc_index, slc_index + 1, 1)
        mask_indices = tuple(mask_indices)

        return np.squeeze(self.get_obstruction_mask(cell_centered=cell_centered)[mask_indices])

    def coordinate_to_index(self, coordinate: Tuple[float, ...],
                            dimension: Tuple[Literal[0, 1, 2, 'x', 'y', 'z'], ...] = (
                                    'x', 'y', 'z'), cell_centered=False) -> Tuple[int, ...]:
        """Finds the nearest point in the mesh's grid and returns its indices.

        :param coordinate: Tuple of 3 floats. If the dimension parameter is supplied, up to 2
            dimensions can be left out from the tuple.
        :param dimension: The dimensions in which to return the indices.
        """
        # Convert possible integer input to chars
        dimension = tuple(('x', 'y', 'z')[dim] if type(dim) == int else dim for dim in dimension)

        ret = list()
        for i, dim in enumerate(dimension):
            co = coordinate[i]
            coords = self.coordinates[dim]
            if cell_centered:
                coords = coords[:-1] + (coords[1] - coords[0]) / 2
            idx = np.searchsorted(coords, co, side="left")
            if co > 0 and (idx == len(coords) or np.math.fabs(co - coords[idx - 1]) < np.math.fabs(
                    co - coords[idx])):
                ret.append(idx - 1)
            else:
                ret.append(idx)
        return tuple(ret)

    def __getitem__(self, dimension: Literal[0, 1, 2, 'x', 'y', 'z']) -> np.ndarray:
        """Get all values in given dimension.

        :param dimension: The dimension in which to return all grid values (0=x, 1=y, 2=z).
        """
        # Convert possible integer input to chars
        if type(dimension) == int:
            dimension = ('x', 'y', 'z')[dimension]
        return self.coordinates[dimension]

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self, *args, **kwargs):
        return f'Mesh(id="{self.id}", extent={str(self.extent)}, dimension={str(self.dimension)})'



from typing import Sequence, BinaryIO, Union, Tuple
import numpy as np

from fdsreader.settings import FORTRAN_DATA_TYPE_CHAR, FORTRAN_DATA_TYPE_FLOAT, FORTRAN_DATA_TYPE_INTEGER, \
    FORTRAN_BACKWARD

_BASE_FORMAT = f"{FORTRAN_DATA_TYPE_INTEGER}, {{}}" + (
    f", {FORTRAN_DATA_TYPE_INTEGER}" if FORTRAN_BACKWARD else "")

_DATA_TYPES = {'i': FORTRAN_DATA_TYPE_INTEGER, 'f': FORTRAN_DATA_TYPE_FLOAT,
               'c': FORTRAN_DATA_TYPE_CHAR, '{}': "{}"}


def _get_dtype_output_format(d, n):
    """Returns the correct output format needed to create a numpy dtype depending on input.
    """
    if d == 'c':
        return str(n)
    if type(n) == int or type(n) == np.int32:
        return f"({n},)"
    return str(n)


def new_raw(data_structure: Sequence[Tuple[str, Union[int, str]]]) -> str:
    """Creates the string definition for a fortran-compliant numpy dtype to read in binary fortran data.

    :param data_structure: Tuple consisting of tuples with 2 elements each where the first element
     is a char ('i', 'f', 'c' or '{}') representing the primitive data type to be used and the
     second element an integer representing the number of times this data type was written out in
     Fortran.
    :returns: The definition string for a fortran-compliant numpy dtype with the desired structure.
    """
    return ", ".join([_get_dtype_output_format(d, n) + _DATA_TYPES[d] for d, n in data_structure])


def new(data_structure: Sequence[Tuple[str, Union[int, str]]]) -> np.dtype:
    """Creates a fortran-compliant numpy dtype to read in binary fortran data.

    :param data_structure: Tuple consisting of tuples with 2 elements each where the first element
     is a char ('i', 'f' or 'c') representing the primitive data type to be used and the second
     element an integer representing the number of times this data type was written out in Fortran.
    :returns: The newly created fortran-compliant numpy dtype with the desired structure.
    """
    return np.dtype(_BASE_FORMAT.format(new_raw(data_structure)))


def combine(*dtypes: np.dtype):
    """Combines multiple numpy dtypes into one.

    :param dtypes: An arbitrary amount of numpy dtype objects can be provided.
    :returns: The newly created numpy dtype.
    """
    count = 0
    type_combination = list()
    for types in dtypes:
        for dtype in types.descr:
            type_combination.append(tuple(['f' + str(count)] + list(dtype[1:])))
            count += 1
    return np.dtype(type_combination)


# Commonly used datatypes
CHAR = new((('c', 1),))
INT = new((('i', 1),))
FLOAT = new((('f', 1),))
# Border datatype to get the border of a fortran write
PRE_BORDER = np.dtype(FORTRAN_DATA_TYPE_INTEGER)
HAS_POST_BORDER = FORTRAN_BACKWARD


def read(infile: BinaryIO, dtype: np.dtype, n: int):
    """Convenience function to read in binary data from a file using a numpy dtype.

    :param infile: Already opened binary IO stream.
    :param dtype: Numpy dtype object.
    :param n: The number of times a dtype object should be read in from the stream.
    :returns: Read in data.
    """
    return np.array(
        [[t[i] for i in range(1, len(t), 3)] for t in np.fromfile(infile, dtype=dtype, count=n)],
        dtype=object)



import os

import setuptools

base_dir = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(base_dir, "README.md"), 'r') as f:
    long_description = f.read()

setuptools.setup(
    name="fdsreader",
    use_incremental=True,
    setup_requires=['incremental'],
    author="FZJ IAS-7 (Prof. Dr. Lukas Arnold, Jan Vogelsang)",
    author_email="j.vogelsang@fz-juelich.de",
    description="Python reader for data generated by FDS.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JanVogelsang/fdsreader",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'incremental',
        'numpy',
        'typing_extensions',
        'idna>=2.5,<3'
    ],
)



import os
from typing import List, Dict, Tuple, Union
from typing_extensions import Literal
import numpy as np

from fdsreader.utils import Surface, Mesh, Extent, Quantity, Dimension
import fdsreader.utils.fortran_data as fdtype
from fdsreader import settings


class Hole:
    """Represents a hole in an obstruction. Used to output default values instead of actual data.

    :ivar dimension: :class:`Dimension` object containing information about steps in each dimension.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar orientation: The direction the patch is facing (x={-1;1}, y={-2;2}, z={-3;3}).
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    """
    default_value = 0

    def __init__(self, dimension: Dimension, extent: Extent, orientation: int, cell_centered: bool):
        self.dimension = dimension
        self.extent = extent
        self.orientation = orientation
        self.cell_centered = cell_centered

    @property
    def shape(self) -> Tuple:
        """Convenience function to calculate the shape of the array containing data for this hole.
        """
        return self.dimension.shape(self.cell_centered)

    @property
    def size(self) -> int:
        """Convenience function to calculate the number of data points in the array for this hole.
        """
        return self.dimension.size(self.cell_centered)

    @property
    def data(self):
        """Creates an array filled with the specified default value (self.default_value).
        """
        return np.full(self.shape, self.default_value)

    def clear_cache(self):
        """Exists only to be consistent with the :class:`Patch` class.
        """
        pass

    def __repr__(self, *args, **kwargs):
        return f"Hole(shape={self.shape}, orientation={self.orientation}, extent={self.extent})"


class Patch:
    """Container for the actual data which is stored as rectangular plane with specific orientation
        and extent.

    :ivar dimension: :class:`Dimension` object containing information about steps in each dimension.
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    :ivar orientation: The direction the patch is facing (x={-1;1}, y={-2;2}, z={-3;3}).
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar n_t: Total number of time steps for which output data has been written.
    """

    def __init__(self, file_path: str, dimension: Dimension, extent: Extent, orientation: int,
                 cell_centered: bool, patch_offset:int, initial_offset: int, n_t: int):
        self.file_path = file_path
        self.dimension = dimension
        self.extent = extent
        self.orientation = orientation
        self.cell_centered = cell_centered
        self._patch_offset = patch_offset
        self._initial_offset = initial_offset
        self._time_offset = -1
        self.n_t = n_t

    @property
    def shape(self) -> Tuple:
        """Convenience function to calculate the shape of the array containing data for this patch.
        """
        return self.dimension.shape(self.cell_centered)

    @property
    def size(self) -> int:
        """Convenience function to calculate the number of data points in the array for this patch.
        """
        return self.dimension.size(self.cell_centered)

    def _post_init(self, time_offset: int):
        """Fully initialize the patch as soon as the number of timesteps is known.
        """
        self._time_offset = time_offset

    @property
    def data(self):
        """Method to load the quantity data for a single patch.
        """
        if not hasattr(self, "_data"):
            dtype_data = fdtype.new((('f', self.dimension.size(cell_centered=False)),))

            if self.n_t == -1:
                self.n_t = (os.stat(self.file_path).st_size - self._initial_offset) // self._time_offset

            self._data = np.empty((self.n_t,) + self.shape)
            with open(self.file_path, 'rb') as infile:
                for t in range(self.n_t):
                    infile.seek(self._initial_offset + self._patch_offset + t * self._time_offset)
                    data = np.fromfile(infile, dtype_data, 1)[0][1].reshape(
                        self.dimension.shape(cell_centered=False), order='F')
                    if self.cell_centered:
                        self._data[t, :] = data[:-1, :-1]
                    else:
                        self._data[t, :] = data
        return self._data

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        if hasattr(self, "_data"):
            del self._data

    def __repr__(self, *args, **kwargs):
        return f"Patch(shape={self.shape}, orientation={self.orientation}, extent={self.extent})"


class Boundary:
    """Container for the actual data which is stored as rectangular plane with specific orientation
        and extent.

    :ivar quantity: Quantity object containing information about the quantity calculated for this
        :class:`Obstruction` with the corresponding label and unit.
    :ivar times: Numpy array containing all times for which data has been recorded.
    :ivar cell_centered: Indicates whether centered positioning for data is used.
    :ivar lower_bounds: Dictionary with lower bounds for each timestep with meshes as keys.
    :ivar upper_bounds: Dictionary with upper bounds for each timestep with meshes as keys.
    :ivar n_t: Total number of time steps for which output data has been written.
    """

    def __init__(self, cell_centered: bool, quantity: Quantity, times: np.ndarray, n_t: int):
        self.cell_centered = cell_centered
        self.quantity = quantity
        self.times = times
        self.lower_bounds: Dict[Mesh, np.ndarray] = dict()
        self.upper_bounds: Dict[Mesh, np.ndarray] = dict()
        self.n_t = n_t
        self.extent = None

        self._patches: Dict[Mesh, List[Union[Patch, Hole]]] = dict()

    def _add_patches(self, mesh: Mesh, patches: List[Union[Patch, Hole]], lower_bounds: np.ndarray,
                     upper_bounds: np.ndarray):
        self._patches[mesh] = patches
        self.lower_bounds[mesh] = lower_bounds
        self.upper_bounds[mesh] = upper_bounds

    @staticmethod
    def sort_patches_cartesian(patches_in: List[Union[Patch, Hole]]):
        """Returns all patches (of same orientation!) sorted in cartesian coordinates.
        """
        patches = patches_in.copy()
        if len(patches) != 0:
            patches_cart = [[patches[0]]]
            orientation = abs(patches[0].orientation)
            if orientation == 1:  # x
                patches.sort(key=lambda p: (p.extent.y_start, p.extent.z_start))
            elif orientation == 2:  # y
                patches.sort(key=lambda p: (p.extent.x_start, p.extent.z_start))
            elif orientation == 3:  # z
                patches.sort(key=lambda p: (p.extent.x_start, p.extent.y_start))

            if orientation == 1:
                for patch in patches[1:]:
                    if patch.extent.y_start == patches_cart[-1][-1].extent.y_start:
                        patches_cart[-1].append(patch)
                    else:
                        patches_cart.append([patch])
            else:
                for patch in patches[1:]:
                    if patch.extent.x_start == patches_cart[-1][-1].extent.x_start:
                        patches_cart[-1].append(patch)
                    else:
                        patches_cart.append([patch])
            return patches_cart
        return patches

    def get_patches_in_mesh(self, mesh: Mesh) -> List[Union[Patch, Hole]]:
        """Gets all patches in a specific mesh.
        """
        if not hasattr(self._patches[mesh][0], "_data"):
            for patch in self._patches[mesh]:
                _ = patch.data
        return self._patches[mesh]

    @property
    def faces(self) -> Dict[Literal[-3, -2, -1, 1, 2, 3], np.ndarray]:
        """Global/combines faces for all 6 orientations.
        """
        if not hasattr(self, "_faces"):
            self._prepare_faces()
        return self._faces

    @property
    def vmin(self) -> float:
        """Minimum value of all patches at any time.
        """
        curr_min = min(np.min(b) for b in self.lower_bounds.values())
        if curr_min == 0.0:
            return min(min(np.min(p.data) for p in ps) for ps in self._patches.values())
        return curr_min

    @property
    def vmax(self) -> float:
        """Maximum value of all patches at any time.
        """
        curr_max = max(np.max(b) for b in self.upper_bounds.values())
        if curr_max == np.float32(-1e33):
            return max(max(np.max(p.data) for p in ps) for ps in self._patches.values())
        return curr_max

    def _prepare_faces(self):
        patches_for_face = {-3: list(), -2: list(), -1: list(), 1: list(), 2: list(), 3: list()}
        for patches in self._patches.values():
            for patch in patches:
                patches_for_face[patch.orientation].append(patch)

        self._faces: Dict[Literal[-3, -2, -1, 1, 2, 3], np.ndarray] = dict()
        for face in (-3, -2, -1, 1, 2, 3):
            patches = self.sort_patches_cartesian(patches_for_face[face])
            if len(patches) == 0:
                continue

            shape_dim1 = sum([patch_row[0].shape[0] for patch_row in patches])
            shape_dim2 = sum([patch.shape[1] for patch in patches[0]])

            self._faces[face] = np.empty(shape=(self.n_t, shape_dim1, shape_dim2))
            dim1_pos = 0
            dim2_pos = 0
            for patch_row in patches:
                d1 = patch_row[0].shape[0]
                for patch in patch_row:
                    d2 = patch.shape[1]
                    self._faces[face][:, dim1_pos:dim1_pos + d1,
                    dim2_pos:dim2_pos + d2] = patch.data
                    dim2_pos += d2
                dim1_pos += d1
                dim2_pos = 0

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for patches in self._patches.values():
            for patch in patches:
                patch.clear_cache()

    def __getitem__(self, item):
        """Either gets all patches in a mesh [type(item)==Mesh] or all faces with a specific
            orientation [type(item)==int].
        """
        if type(item) == Mesh:
            return self.get_patches_in_mesh(item)
        else:
            return self.faces[item]

    def __repr__(self, *args, **kwargs):
        return f"Boundary(quantity={self.quantity}, cell_centered={self.cell_centered})"


class Obstruction:
    """A box-shaped obstruction with specific surfaces (materials) on each side.

    :ivar id: ID of the obstruction.
    :ivar side_surfaces: Tuple of six surfaces for each side of the cuboid.
    :ivar bound_indices: Indices used to define obstruction bounds in terms of mesh locations.
    :ivar color_index: Type of coloring used to color obstruction.
     \n-1 - default color
     \n-2 - invisible
     \n-3 - use red, green, blue and alpha (rgba attribute)
     \nn>0 - use n’th color table entry
    :ivar block_type: Defines how the obstruction is drawn.
     \n-1 - use surface to obtain blocktype
     \n0 - regular block
     \n2 - outline
    :ivar texture_origin: Origin position of the texture provided by the surface. When the texture
        does have a pattern, for example windows or bricks, the texture_origin specifies where the
        pattern should begin.
    :ivar rgba: Optional color of the obstruction in form of a 4-element tuple
        (ranging from 0.0 to 1.0).
    :ivar extent: :class:`Extent` object containing 3-dimensional extent information.
    """

    def __init__(self, oid: int,
                 side_surfaces: Tuple[Surface, Surface, Surface, Surface, Surface, Surface],
                 bound_indices: Tuple[int, int, int, int, int, int], color_index: int,
                 block_type: int, texture_origin: Tuple[float, float, float],
                 rgba: Union[Tuple[()], Tuple[float, float, float, float]] = ()):
        self.id = oid
        self.side_surfaces = side_surfaces
        self.bound_indices = {'x': (bound_indices[0], bound_indices[1]),
                              'y': (bound_indices[2], bound_indices[3]),
                              'z': (bound_indices[4], bound_indices[5])}
        self.color_index = color_index
        self.block_type = block_type
        self.texture_origin = texture_origin
        if len(rgba) != 0:
            self.rgba = rgba

        self._extents: Dict[Mesh, Union[Extent, List[Extent]]] = dict()
        self.extent: Extent = tuple()

        self._boundary_data: Dict[int, Boundary] = dict()

    def _post_init(self):
        vals = list()
        for val in self._extents.values():
            if type(val) == list:
                for v in val:
                    vals.append(v)
            else:
                vals.append(val)

        self.extent = Extent(
            min(vals, key=lambda e: e.x_start).x_start, max(vals, key=lambda e: e.x_end).x_end,
            min(vals, key=lambda e: e.y_start).y_start, max(vals, key=lambda e: e.y_end).y_end,
            min(vals, key=lambda e: e.z_start).z_start, max(vals, key=lambda e: e.z_end).z_end)
        for boundary in self._boundary_data.values():
            boundary.extent = self.extent

    def _add_patches(self, bid: int, cell_centered: bool, quantity: str, label: str, unit: str,
                     mesh: Mesh, patches: List[Union[Patch, Hole]], times: np.ndarray, n_t: int,
                     lower_bounds: np.ndarray, upper_bounds: np.ndarray):
        if bid not in self._boundary_data:
            self._boundary_data[bid] = Boundary(cell_centered, Quantity(quantity, label, unit),
                                                times, n_t)
        self._boundary_data[bid]._add_patches(mesh, patches, lower_bounds, upper_bounds)

        if not settings.LAZY_LOAD:
            self._boundary_data[bid].get_patches_in_mesh(mesh)

    @property
    def quantities(self) -> List[Quantity]:
        """Get a list of all quantities for which boundary data exists.
        """
        return [b.quantity for b in self._boundary_data.values()]

    def get_boundary_data(self, quantity: Union[Quantity, str]):
        """Gets the boundary data for a specific quantity.
        """
        if type(quantity) != str:
            quantity = quantity.quantity
        return next((x for x in self._boundary_data.values() if
                     x.quantity.quantity.lower() == quantity.lower() or
                     x.quantity.label.lower() == quantity.lower()), None)

    @property
    def has_boundary_data(self):
        """Whether boundary data has been output in the simulation.
        """
        return len(self._boundary_data) != 0

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for bndf in self._boundary_data.values():
            bndf.clear_cache()

    def __getitem__(self, item):
        if type(item) == Quantity or type(item) == str:
            return self.get_boundary_data(item)
        return self._boundary_data[item]

    def __eq__(self, other):
        return self.id == other.id and self.extent == other.extent

    def __repr__(self, *args, **kwargs):
        return f"Obstruction(id={self.id}, extent={self.extent}" + \
               (f", Quantities={self.quantities}" if self.has_boundary_data else "") + ")"
